--************************************************************************************************************************************************************	
--move the update of the wheelshape to the "postupdate" function so that it has the lastest values
Attachable.mrUpdate = function(self, superFunc, dt)

	if not self.mrIsMrVehicle then
		return superFunc(self, dt);
	end;
	
	if self:getIsActive() then
		if self.updateSteeringAxleAngle then
			local baseVehicle = self:getSteeringAxleBaseVehicle()
			if baseVehicle ~= nil and (self.movingDirection >= 0 or self.steeringAxleUpdateBackwards) then	
				if self:getLastSpeed()>self.steeringAxleAngleScaleEnd then
					self.steeringAxleTargetAngle = 0;
				else
					local yRot = Utils.getYRotationBetweenNodes(self.steeringAxleNode, baseVehicle.steeringAxleNode);
					--if math.abs(yRot) > 1.57 then
					if math.abs(yRot) > 2.356 then -- allow more than 90 angle between trailer and tractor (example : semi trailer)
						if yRot > 0 then
							yRot = yRot - 3.14;
						else
							yRot = yRot + 3.14;
						end;
					end;
					local startSpeed = self.steeringAxleAngleScaleStart;
					local endSpeed = self.steeringAxleAngleScaleEnd;
					local scale = Utils.clamp(1 + (self:getLastSpeed()-startSpeed)/(startSpeed-endSpeed),0, 1);
					self.steeringAxleTargetAngle = yRot*scale;
				end
			elseif self:getLastSpeed() > 0.2 then
				self.steeringAxleTargetAngle = 0;
			end

			local dir = Utils.sign(self.steeringAxleTargetAngle-self.steeringAxleAngle);
			if dir == 1 then
				self.steeringAxleAngle = math.min(self.steeringAxleAngle + dir*dt*self.steeringAxleAngleSpeed, self.steeringAxleTargetAngle);
			else
				self.steeringAxleAngle = math.max(self.steeringAxleAngle + dir*dt*self.steeringAxleAngleSpeed, self.steeringAxleTargetAngle);
			end
		end
	end
	
end
Attachable.update = Utils.overwrittenFunction(Attachable.update, Attachable.mrUpdate);


--************************************************************************************************************************************************************	
--correct brakeForce function of wheel radius
Attachable.postUpdate = function(self, dt)
	
	if self.mrIsMrVehicle and self:getIsActive() then
		if self.firstTimeRun and self.updateWheels and self.isServer and self.isAddedToPhysics then		
			--print(tostring(g_currentMission.time) .. " postUpdate for : " .. self.configFileName .. " - debugTime=" .. tostring(self.debugTime));
			Attachable.mrBrakeWheels(self, self.brakeForce*self.brakePedal, false);				
		end
		
		--if not self.debugTime then self.debugTime = 0 end;
		--self.debugTime = self.debugTime + dt;
		
	end

end



--************************************************************************************************************************************************************	
--call "mrBrakeWheels"
Attachable.mrOnBrake = function(self, superFunc, brakePedal, forced)
	
	if not self.mrIsMrVehicle then
		return superFunc(self, brakePedal, forced);
	end;
	
	--print(tostring(g_currentMission.time) .. " onbrake for "..self.configFileName .. " - brakePedal="..tostring(brakePedal) .. " - forced="..tostring(forced).. " - debugTime=" .. tostring(self.debugTime));

	--do not brake for the first 2 seconds after attaching, to allow balancing of the difference between the attacher joints
	--20170121 not recommended when loading a game. Indeed, the trailer/tractor set can be parked in a slope and so, if the trailer does not brake when loaded, the tractor + trailer set can roll down the slope during 2s	
	if self.attachTime==0 or self.attachTime+2000 < g_currentMission.time or forced then
		self.brakePedal = brakePedal;		
		
		if self.isServer and self.isAddedToPhysics and forced then --20170315 - only call mrBrakeWheels	if "forced==true" (otherwise, this is the "Attachable.mrUpdate" function that takes care of that
			Attachable.mrBrakeWheels(self, self.brakeForce*brakePedal, forced);
		end
		for _,implement in pairs(self.attachedImplements) do
			if implement.object ~= nil then
				implement.object:onBrake(brakePedal, forced);
			end
		end
	end
end
Attachable.onBrake = Utils.overwrittenFunction(Attachable.onBrake, Attachable.mrOnBrake);

--************************************************************************************************************************************************************	
--call "mrBrakeWheels"
Attachable.mrOnReleaseBrake = function(self, superFunc)
	
	if not self.mrIsMrVehicle then
		return superFunc(self);
	end;
	
	--print(tostring(g_currentMission.time) .. " onreleasebrake for "..self.configFileName.. " - debugTime=" .. tostring(self.debugTime));

	self.brakePedal = 0;
	--20170315 - no need to call the "mrBrakeWheels" function since the "Attachable.mrUpdate" already takes care of that
	--if self.isServer and self.isAddedToPhysics then
		--for _,wheel in pairs(self.wheels) do			
			--setWheelShapeProps(wheel.node, wheel.wheelShape, 0, 0, wheel.steeringAngle, wheel.rotationDamping);
			--Attachable.mrBrakeWheels(self, 0, false);
		--end
	--end
	for _,implement in pairs(self.attachedImplements) do
		if implement.object ~= nil then
			implement.object:onReleaseBrake();
		end
	end
end
Attachable.onReleaseBrake = Utils.overwrittenFunction(Attachable.onReleaseBrake, Attachable.mrOnReleaseBrake);


--************************************************************************************************************************************************************	
--correct brakeForce function of wheel radius
Attachable.mrBrakeWheels = function(self, brakeForce, forced)

	--forced == true means the tool is detached from the tractor and so, "mrBrakeWheels" will not be called again until the tool is attached to a tractor
	
	--print(tostring(g_currentMission.time) .. " - mrBrakeWheels - brakeForce=" .. tostring(brakeForce) .. " - forced=" .. tostring(forced) .. " for : " .. self.configFileName);

	for _,wheel in pairs(self.wheels) do --need to brake when detaching an implement otherwise, since the implement is not active anymore, the "update" function does not brake the wheels.		
	
		if wheel.mrNotAWheel then
			if forced then
				setWheelShapeProps(wheel.node, wheel.wheelShape, 0, 0.1, 0, 0.1 * wheel.mass);--wheel.rotationDamping);
			else
				--print(" mrNotAWheel - forced = false, hasGroundContact="..tostring(wheel.hasGroundContact).." - mass="..tostring(wheel.mass) .. " - damping="..tostring(wheel.rotationDamping));
				local damping = wheel.rotationDamping;
				if not wheel.hasGroundContact and damping==0 then					
					damping = 0.03 * wheel.mass;							
					--print("------------- new damping="..tostring(damping));
				end
				setWheelShapeProps(wheel.node, wheel.wheelShape, 0, 0, 0, damping);--wheel.rotationDamping);	
			end
		else
	
			local rotDamping = wheel.rotationDamping;			
			local finalBrakeForce = brakeForce*wheel.brakeFactor*wheel.radius;
			
			--print("test finalbrakeForce = " .. tostring(finalBrakeForce));
			
			if finalBrakeForce>0 then			
				--20170308 - limit brake force according to weight on the wheel
				--finalBrakeForce = Utils.clamp(1.1*wheel.mrLastTireLoad*wheel.radius/math.max(1, wheel.mrLastFrictionFx)^0.5, 0.1*finalBrakeForce, finalBrakeForce);
				if not forced then
					finalBrakeForce = Utils.clamp(1.1*wheel.mrLastTireLoad*wheel.radius/Utils.clamp(wheel.mrLastFrictionFx, 1, 1.6), 0.1*finalBrakeForce, finalBrakeForce);	
				end
			else -- final brake force == 0
				--print("test 2 - finalBrakeForce==0 but hasGroundContact="..tostring(wheel.hasGroundContact));
				if forced then --implement detached from the tractor, adding some "rolling resistance" to fake wheels like cultivator rollbars or weeder tines ?
					--20170305 - add damping to wheels to avoid "infinite" rolling of the attachable when detached (useful for "attachable" with all the weight on wheels that are not braked)						
					finalBrakeForce = 2*wheel.radius;
					--print("test - rotdamping - wheel mass="..tostring(wheel.mass) .. " - finalBrakeForce=" .. tostring(finalBrakeForce));
				elseif not wheel.hasGroundContact and rotDamping==0 then
					rotDamping = 0.03 * wheel.mass; --20170215 - takes into account wheel mass (same time to reach 0 rotSpeed whatever the mass of the wheel)					
				end		
			end
			
			
			
			--simulate rolling resistance at still
			--[[ 2017 01 10 no need since we are using the small steeringAngle to stabilize lateral slip of wheels
			if finalBrakeForce==0 then
				local wheelGroundSpd = math.abs(wheel.mrLastWheelSpeed*wheel.radius);
				if wheelGroundSpd<0.2 then --0.2 = 0.72kph
					--finalBrakeForce = wheel.mrLastTireLoad * wheel.mrCurrentRollingResistance*wheel.radius*math.min(1, 2-10*wheelGroundSpd);
					rotDamping = rotDamping + 1*wheel.mrCurrentRollingResistance*math.min(1, 1-5*wheelGroundSpd)/wheel.radius;
				end
				
			--else
				--"remove" the mrWheelsFrictionFactor benefit from the braking force
				--finalBrakeForce = finalBrakeForce*wheel.brakeFactor*wheel.radius / self.mrWheelsFrictionFactor;
			end
			--]]
			
			
			--when running at relatively high speed, if all wheels are aligned (tractor and trailer), the lateral slip should be 0, but in such a case, the lateral force would be 0 too (like driving on ice)
			--and so, the base game engine uses a value of 0.35 or -0.35 when the lateral slip should be 0
			--this is a major problem since if the slip becomes 0.01, it can keep going from 0.01 to 0.35 which means poor lateral friction/great lateral friction.
			--when the trailer is featured with 4 wheels or more, this can (and will) cause (especially visible with high load and high speed) the trailer to rotate against its y axis, making the tractor backward axle sustaining a great lateral force (if the tractor is too light, it will slip laterally on its backward axle)
			local steeringAngle = wheel.steeringAngle;
			--[[if wheel.steeringAxleScale==0 and steeringAngle==0 and not wheel.mrNotAWheel then		
				if wheel.positionX>0 then				
					steeringAngle = 0;--0.03;				
				else				
					steeringAngle = 0;---0.03;				
				end
			end--]]
			if steeringAngle==0 then
				steeringAngle = wheel.mrStabilizingSteeringAngle;
			end
			
			setWheelShapeProps(wheel.node, wheel.wheelShape, 0, finalBrakeForce, steeringAngle, rotDamping);
			
		end
		
	end
end



--MR : fix bug with upperRotationOffset not "zero" for trailer/semitrailer
Attachable.mrLoadInputAttacherJoint = function(self, superFunc1, superFunc, xmlFile, key, inputAttacherJoint, index)	

	local result = superFunc1(self, superFunc, xmlFile, key, inputAttacherJoint, index);

	if result==true then
		--print("test - mr vehicle = " .. tostring(self.mrIsMrVehicle) .. " - hard attch=" .. tostring(inputAttacherJoint.hardAttach));
		if self.mrIsMrVehicle and not inputAttacherJoint.hardAttach then
			--we want default value for inputAttacherJoint.upperRotationOffset = 0 for joint_type = trailer, trailerlow or semitrailer			
			if inputAttacherJoint.jointType == AttacherJoints.JOINTTYPE_TRAILER or inputAttacherJoint.jointType == AttacherJoints.JOINTTYPE_TRAILERLOW  or inputAttacherJoint.jointType == AttacherJoints.JOINTTYPE_SEMITRAILER then
				inputAttacherJoint.upperRotationOffset = math.rad(Utils.getNoNil(getXMLFloat(xmlFile, key .. "#upperRotationOffset"), 0));
			end			
		end
	end
	
	return result;
	
end
Attachable.loadInputAttacherJoint = Utils.overwrittenFunction(Attachable.loadInputAttacherJoint, Attachable.mrLoadInputAttacherJoint);
