--************************************************************************************************************************************************************	
--MR : load the forceNodeCOM parameter if present
PowerConsumer.mrLoadPowerSetup = function(self, xmlFile)

	if self.powerConsumer.forceNode then
		local x, y, z = Utils.getVectorFromString(getXMLString(xmlFile, "vehicle.powerConsumer#forceNodeCOM"));
		if x == nil or y == nil or z == nil then
			x,y,z = getCenterOfMass(self.powerConsumer.forceNode);		
		end
		self.powerConsumer.forceNodeCOM = {};
		self.powerConsumer.forceNodeCOM.x = x;
		self.powerConsumer.forceNodeCOM.y = y;
		self.powerConsumer.forceNodeCOM.z = z;	
		
		self.powerConsumer.mrExistingForcePerLoadedTon = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.powerConsumer#mrExistingForcePerLoadedTon"), 0);
		
		local workingWidthAnimationName = getXMLString(xmlFile, "vehicle.powerConsumer.mrWorkingWidthAnimation#animName");
		if workingWidthAnimationName then
			self.powerConsumer.mrWorkingWidthAnimation = {};
			self.powerConsumer.mrWorkingWidthAnimation.animName = workingWidthAnimationName;
			self.powerConsumer.mrWorkingWidthAnimation.startForceFactor = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.powerConsumer.mrWorkingWidthAnimation#startForceFactor"),1);
			self.powerConsumer.mrWorkingWidthAnimation.endForceFactor = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.powerConsumer.mrWorkingWidthAnimation#endForceFactor"),1);
		end
		
	end;
	
	self.powerConsumer.mrLastForceApplied = 0;
	self.mrAreaDependantPtoPower = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.powerConsumer#areaDependantPtoPower"), 0);
	self.powerConsumer.mrLoweredMaxForce = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.powerConsumer#loweredMaxForce"), 0);
	
	
	
	--self.mrAvgCutterArea = 0;
	


end
PowerConsumer.loadPowerSetup = Utils.appendedFunction(PowerConsumer.loadPowerSetup, PowerConsumer.mrLoadPowerSetup);



--************************************************************************************************************************************************************	
--MR : a little more force as the speed increase to help tractor stabilize speed and better reflect reality
PowerConsumer.mrUpdate = function(self, superFunc, dt)
	
	if not self.mrIsMrVehicle then
		return superFunc(self, dt);
	end;	

	if self.isServer and (0<self.powerConsumer.maxForce or 0<self.powerConsumer.mrLoweredMaxForce) then
	
		local force = 0;
		local maxForce = self.powerConsumer.maxForce;
		if self:isLowered(false) then
			maxForce = maxForce + self.powerConsumer.mrLoweredMaxForce;
		end
		if self:getIsActive() and maxForce>0 then		
	
			if Vehicle.debugRendering then
				self.mrDebugPowerConsumerLastForce = 0;
			end						
			
			local spd = self.lastSpeedReal*1000;
			if self.attacherVehicle and spd<0.6 then -- 0.6 = 2.16kph
				spd = self.attacherVehicle.lastSpeedReal*1000;
			end
			
			if self.powerConsumer.forceNode ~= nil and spd>0.03 then	--0.03 = 0.108 km/h
				local multiplier = self:getPowerMultiplier();
				if multiplier ~= 0 then					
				
					local vx, vy, vz = getLinearVelocity(self.powerConsumer.forceNode);
					
					if vx==nil or vy==nil or vz==nil then				
						RealisticUtils.printWarning("PowerConsumer.mrUpdate", "getLinearVelocity has returned a nil value. Check 'forceDirNode' for " .. tostring(self.configFileName), true);				
					else
						
						--check animation working width force factor
						if self.powerConsumer.mrWorkingWidthAnimation then
							local currentAnimTime = self:getAnimationTime(self.powerConsumer.mrWorkingWidthAnimation.animName);
							maxForce = maxForce * Utils.lerp(self.powerConsumer.mrWorkingWidthAnimation.startForceFactor, self.powerConsumer.mrWorkingWidthAnimation.endForceFactor, currentAnimTime);							
							--print("currentWidthAnimTime = "..tostring(currentAnimTime) .. " - fx = " .. tostring(Utils.lerp(self.powerConsumer.mrWorkingWidthAnimation.startForceFactor, self.powerConsumer.mrWorkingWidthAnimation.endForceFactor, currentAnimTime)));
						end						
						
						if self.currentMass and self.emptyMass then
							maxForce = maxForce - self.powerConsumer.mrExistingForcePerLoadedTon * (self.currentMass-self.emptyMass);
						end
					
						--problem : when lowering the tool (tractor at still), the powerconsumerForce is applied to the tool (direction mostly upward since the tool is going down)
						--this can cause weird behavior and visual
						--solution : only apply the resistance force when the "3pt hitch" is not in movement
						
						local frictionForce = self.powerConsumer.forceFactor * spd * self.powerConsumer.totalMass / (dt/1000);
						if self.movingDirection ~= self.powerConsumer.forceDir then
							--apply less force in reverse (tool not working)							
							force = 0.25 * maxForce * spd;
						else						
							--frictionForce = self.powerConsumer.forceFactor * spd * 3.6 * self.powerConsumer.maxForce; -- max force at 1kph if forceFactor==1
							force = maxForce * (1 + 0.03*spd); --10% more draft at 12kph					
							
							if Vehicle.debugRendering then
								if PowerConsumer.mrDebugForce then
									force = PowerConsumer.mrDebugForce;
								end
							end
							
						end						
						force = math.min(frictionForce, force);	

						if force>self.powerConsumer.mrLastForceApplied then
							force = math.min(force, self.powerConsumer.mrLastForceApplied + maxForce * dt/1500); -- 1.5s to get maxForce
							--force = self.powerConsumer.mrLastForceApplied * 0.99 + force * 0.01;
						end
					
						local spd = Utils.vector3Length(vx, vy, vz);						
						
						addForce(self.powerConsumer.forceNode, -force*vx/spd,-force*vy/spd,-force*vz/spd, self.powerConsumer.forceNodeCOM.x,self.powerConsumer.forceNodeCOM.y,self.powerConsumer.forceNodeCOM.z, true);
						
						if Vehicle.debugRendering then
							self.mrDebugPowerConsumerLastForce = force;
							if self:getIsActiveForInput() then
								local str = string.format("frictionForce=%.2f maxForce=%.2f -> current force=%.2f", frictionForce, maxForce, force);
								renderText(0.68, 0.85, getCorrectTextSize(0.02), str);
							end
						end	
					end -- check vx vy vz
				end -- check multiplier
			end--check powerconsumerforce and speed		

			--if self.attacherVehicle and self.attacherVehicle.isEntered then
			--print(tostring(g_currentMission.time) .. " test force = " .. tostring(force));
			--end
			
		end--check active
		
		self.powerConsumer.mrLastForceApplied = force;
		
		
		
	end
	
end
PowerConsumer.update = Utils.overwrittenFunction(PowerConsumer.update, PowerConsumer.mrUpdate);


--************************************************************************************************************************************************************	
--MR : add power function of "load" of the implement
--for combine/harvester cutters : power function of the "self.lastCutterArea"
PowerConsumer.mrGetConsumedPtoTorque=function(self, superFunc)

	if not self.mrIsMrVehicle then
		return superFunc(self, dt);
	end
	
	if self:getDoConsumePtoPower() then
		local rpm = self.powerConsumer.ptoRpm;
		if rpm > 0.001 then
			local totalPower = self.powerConsumer.neededPtoPower;
			--[[ 20170220 - not useful anymore - directly taken into account into the combine class
			if self.mrAreaDependantPtoPower>0 then
				--combine/harvester cutter
				if self.mrAvgCutterArea>0  then
					totalPower = totalPower + self.mrAreaDependantPtoPower * self.mrAvgCutterArea^0.5;
				end
			end--]]			
			
			return totalPower/(rpm*math.pi/30);
		end
	end	
	
	return 0;
end
PowerConsumer.getConsumedPtoTorque = Utils.overwrittenFunction(PowerConsumer.getConsumedPtoTorque, PowerConsumer.mrGetConsumedPtoTorque);

--************************************************************************************************************************************************************	
--MR : update avgCutterArea -- 20170220 - not useful anymore, mrAreaDependantPtoPower is directly taken into account in the combine class
--[[
PowerConsumer.mrUpdateTick=function(self, dt)
	if not self.mrIsMrVehicle then
		return;
	end
	
	if self.isServer and self.mrAreaDependantPtoPower>0 then
		if self.lastCutterArea then
			if self.reelStarted then				
				self.mrAvgCutterArea = (1-0.0005*dt) * self.mrAvgCutterArea + 0.0005*dt*self.lastCutterArea;				
				--print("test power consumer self.lastCutterArea=" .. tostring(self.lastCutterArea) .. " - self.mrAvgCutterArea="..tostring(self.mrAvgCutterArea));
			else
				self.mrAvgCutterArea = 0;
			end
		end
	end
	
end
PowerConsumer.updateTick = Utils.appendedFunction(PowerConsumer.updateTick, PowerConsumer.mrUpdateTick);--]]





addConsoleCommand("mrsetPowerConsumerForce", "Set the value of PowerConsumer.mrDebugForce", "consoleCommandSetForce", PowerConsumer);
PowerConsumer.consoleCommandSetForce = function(unusedSelf, value)
	if value and tonumber(value) then
		PowerConsumer.mrDebugForce = value;
		return "PowerConsumer.consoleCommandSetForce - value=" .. tostring(value);
	end
	return "PowerConsumer.consoleCommandSetForce - no valid value typed as a parameter";
end







