

--************************************************************************************************************************************************************
--** use mr vehicle xml file instead of the one from the game files
Vehicle.mrLoadFinished = function(self, i3dNode, arguments)

	self.mrIsMrVehicle = false;

	--replace xml file with moreRealistic data xml file if present for this vehicle
	if RealisticUtils.defaultVehiclesModifiedData[self.configFileName] then	
		delete(self.xmlFile);
		self.xmlFile = loadXMLFile("TempConfig", RealisticUtils.defaultVehiclesModifiedData[self.configFileName]);
		self.mrIsMrVehicle = true;		
	end
	
	if hasXMLProperty(self.xmlFile, "vehicle.moreRealistic") then
		local defaultValue = true;	--default value when the "moreRealistic" is present
		self.mrIsMrVehicle = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.moreRealistic#isMrVehicle"), defaultValue);
	end
	
	if self.mrIsMrVehicle then
		self.loadWheelDataFromXML = Vehicle.mrLoadWheelDataFromXML;
		self.updateWheelTireFriction = Vehicle.mrUpdateWheelTireFriction;	
	end
	
end;
Vehicle.loadFinished = Utils.prependedFunction(Vehicle.loadFinished, Vehicle.mrLoadFinished);



--************************************************************************************************************************************************************
--load mr parameters
Vehicle.mrLoadFinished2 = function(self, i3dNode, arguments)

	local dragArea = 5;	
	

	if self.mrIsMrVehicle then
		--try to load moreRealistic data from the xml file
		if RealisticUtils.defaultVehiclesModifiedData[self.configFileName] then	
			self.xmlFile = loadXMLFile("TempConfig", RealisticUtils.defaultVehiclesModifiedData[self.configFileName]);
		else
			self.xmlFile = loadXMLFile("TempConfig", self.configFileName);
		end;
				
		self.mrUseMrTransmission = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.moreRealistic#useMrTransmission"), false);
		
		dragArea = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.moreRealistic.dragArea#value"), dragArea);
		local overridedTrailerArea = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.moreRealistic.overridedTrailerArea#value"), 0);
		
		if self.motor and self.mrUseMrTransmission then
			self.motor.rotInertiaFx = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.moreRealistic.engine#rotInertiaFx"), 1); --the lower, the quicker the axle can gain speed		
			
			--local dampingRateFx = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.moreRealistic.engine#dampingRateFx"), 1);	-- 1 = base scale for an diesel engine
			--self.motor.dampingRate = dampingRateFx * self.motor.mrMaxTorque/225; -- NOTE : final "dampingRate" = rotInertia * dampingRate, so always tune the rotInertia before the dampingRate. NOTE2 : damping not useful anymore
			
			self.motor.mrMinSpeed = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.moreRealistic.engine#minSpeed"), 2)/3.6; --kph in the xml file, m/s in the lua
			self.motor.minSpeed = self.motor.mrMinSpeed*3.6;
			if self.cruiseControl then self.cruiseControl.minSpeed = self.motor.mrMinSpeed*3.6; end;
			
			self.motor.mrTransmissionEfficiency = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.moreRealistic.engine#transmissionEfficiency"), 0.87);			
			self.motor.mrEngineBrakeRpm = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.moreRealistic.engine#engineBrakeRpm"), 250 + self.motor.maxRpm);	
			self.motor.mrAccPedalRegulateTorque = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.moreRealistic.engine#accPedalRegulateTorque"), false);	
			self.motor.mrIdleEngineRpm = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.moreRealistic.engine#idleEngineRpm"), self.motor.minRpm);			
			self.motor.mrBrakingForceFx = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.moreRealistic.engine#engineBrakeForceFx"), 1);
			self.motor.mrTransmissionReactivity  = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.moreRealistic.engine#transmissionReactivity"), 1);
			self.motor.mrMinRpmAllowedWithLoad = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.moreRealistic.engine#minRpmAllowedWithLoad"), self.motor.mrMinRpmAllowedWithLoad);
			
			
			self.motor.mrRpmRange = self.motor.maxRpm - self.motor.mrIdleEngineRpm;
			self.motor.mrEngineBrakingForce = self.motor.mrBrakingForceFx * self.motor.mrMaxPower * 10 / self.motor.mrEngineBrakeRpm;
			
			
			--------------------------------------------------------------------------------------------------------
			-- overwritten function			
			self.motor.mrUpdateVehicleProps = VehicleMotor.mrUpdateVehicleProps;
			--------------------------------------------------------------------------------------------------------
			
			local defaultFuelUsage = 0.2717; -- default fuel usage (in Liter per Hour per KW):  20L/Hour for 100 PTO Hp (73.6KW)
			local fuelUsageFactor = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.moreRealistic.engine#fuelUsageFactor"), 1);
			self.fuelUsage = defaultFuelUsage * fuelUsageFactor / (60*60*1000); -- from l/KW/h to l/KW/ms		

			self.mrLastBrakePedal = 1;
			self.mrSAGBtryInversionFailed = false;
			self.mrSAGBtryInversion = false;
			
		end
		
		--mr parameters			
		self.mrManageWheelsFriction = Vehicle.mrManageWheelsFriction;
		self.mrRollingResistanceS = 0;
		self.mrAvgWheelSlip = 0;		
		self.mrWheelsFrictionFactor = 1;	
		self.mrTotalWeightOnTires = 0;
		self.mrAvgWheelGroundFrictionCoeff = 0;
		self.mrAvgDrivenWheelsSpeed = 0;
		self.mrTotalxDriveDiff = 0; --store the total difference between each update of the wheels graphics
		self.mrVehicleIsStill = false; --updated by wheelUtils.updateWheelsGraphics
		self.mrDragArea = dragArea;	
		self.mrOverridedTrailerArea = overridedTrailerArea;
		self.mrIsBraking = true;
		self.mrLastWantedAcc = 0;
		
		self.mrGenuineSpeedLimit = self.speedLimit;
		
		local wheelNum = 0;
		local i=0;
		for _, wheel in pairs(self.wheels) do
			i = i + 1;
			wheel.mrNum = i;
			wheel.mrLastWheelSpeed = 0;			
			wheel.mrNormalizedTorqueRatio = Vehicle.mrComputeWheelFinalTorqueRatio(self, wheel);
			if wheel.mrNormalizedTorqueRatio~=1 then
				wheelNum = wheelNum + 1;
			end
			wheel.mrTotalContactWidth = Vehicle.mrGetWheelTotalContactWidth(wheel);
		end
		
		self.mrWheelsNum = i;
		
		if wheelNum>0 then
			for _, wheel in pairs(self.wheels) do				
				if wheel.mrNormalizedTorqueRatio~=1 then
					wheel.mrNormalizedTorqueRatio = wheel.mrNormalizedTorqueRatio * wheelNum;
				end
			end
		end
		
		Vehicle.mrUpdateWheelsIsDriven(self);	

		--20170326 - allow us to "unlink" some node at runtime to fix some problem in i3d files (example : horsch maestro, collisions of the markers arm are bound to the main body)
		RealisticUtils.manageNodes(self, self.xmlFile);
		
	
		delete(self.xmlFile);
		self.xmlFile = nil;	
		
	end
	
end;
Vehicle.loadFinished = Utils.appendedFunction(Vehicle.loadFinished, Vehicle.mrLoadFinished2);

--************************************************************************************************************************************************************
-- allow us to specify a different mass for a "component" than what is set in the i3d file
-- load and set "mrSetMass" value
Vehicle.mrLoadComponentFromXML = function(self, superfunc, component, xmlFile, key, rootPosition, i)

	local result = superfunc(self, component, xmlFile, key, rootPosition, i);
	if result then
		local wantedMass = getXMLFloat(xmlFile, key .. "#mrSetMass");
		if wantedMass~=nil and wantedMass~=0 then
			setMass(component.node, wantedMass);
		end
	end
	return true;
	
end
Vehicle.loadComponentFromXML = Utils.overwrittenFunction(Vehicle.loadComponentFromXML, Vehicle.mrLoadComponentFromXML);

--************************************************************************************************************************************************************
--** override the "Vehicle.update" function just to pass the wheel parameter to the "WheelsUtil.getTireFriction" function (called the "mrGetTireFriction" instead since we have to override the vehicle.update function anyway)
Vehicle.mrUpdate = function(self, superfunc, dt)
	
	if not self.mrIsMrVehicle then
		return superfunc(self, dt);
	end;	
	
	
	if not self.isServer and self.positionIsDirty then
		local maxIntrpAlpha = 1.2;
		local interpolationAlpha = self.interpolationAlpha + g_physicsDtUnclamped / self.interpolationDuration;
		--local interpolationAlpha = math.max((g_physicsNetworkTime-g_clientInterpDelay - self.lastPoseTime) / (self.targetPoseTime - self.lastPoseTime), 0);
		--self.curPoseTime = g_physicsNetworkTime-g_clientInterpDelay;
		--[[local index1 = math.max(table.getn(self.networkPosesTimes)-1, 1);
		local index2 = table.getn(self.networkPosesTimes);
		for i=1, table.getn(self.networkPosesTimes) do
			if self.networkPosesTimes[i] >= g_physicsNetworkTime-g_clientInterpDelay then
				index2 = i;
				index1 = math.max(i-1, 1);
				break;
			end
		end
		if index1 ~= index2 then
			self.interpolationAlpha = math.max((g_physicsNetworkTime-g_clientInterpDelay - self.networkPosesTimes[index1]) / (self.networkPosesTimes[index2] - self.networkPosesTimes[index1]), 0);
		else
			self.interpolationAlpha = maxIntrpAlpha;
		end
		if self.isEntered then
			local component = self.components[1];
			local dist = Utils.vector3Length(component.networkPoses[index2][1]-component.networkPoses[index1][1], component.networkPoses[index2][2]-component.networkPoses[index1][2], component.networkPoses[index2][3]-component.networkPoses[index1][3]);
			local dt = (self.networkPosesTimes[index2] - self.networkPosesTimes[index1]);
			print(string.format("Interp %.2f: %d@%.2fms -> %d@%.2fms: %.2f  speed: %.2f in %.2fms", g_physicsNetworkTime-g_clientInterpDelay, index1, self.networkPosesTimes[index1], index2, self.networkPosesTimes[index2], self.interpolationAlpha, dist/(dt*0.001), dt));
		end]]
		if interpolationAlpha >= maxIntrpAlpha then
			interpolationAlpha = maxIntrpAlpha;
			self.positionIsDirty = false;
		end;
		self.interpolationAlpha = interpolationAlpha;
		for i, component in pairs(self.components) do
			local lastTranslation = component.lastTranslation;
			local targetTranslation = component.targetTranslation;
			local x = lastTranslation[1] + interpolationAlpha * (targetTranslation[1] - lastTranslation[1]);
			local y = lastTranslation[2] + interpolationAlpha * (targetTranslation[2] - lastTranslation[2]);
			local z = lastTranslation[3] + interpolationAlpha * (targetTranslation[3] - lastTranslation[3]);
			local rot1 = component.lastRotation;
			local rot2 = component.targetRotation;
			local qx,qy,qz,qw = Utils.nlerpQuaternionShortestPath(rot1[1], rot1[2], rot1[3], rot1[4], rot2[1], rot2[2], rot2[3], rot2[4], interpolationAlpha);
			self:setWorldPositionQuaternion(x,y,z, qx,qy,qz,qw, i, false);
		end;
	end;

	self.isActive = self:getIsActive();

	self.updateLoopIndex = g_updateLoopIndex;

	if self.isServer then
		--20170318 - add check up to be more reliable in network games (protection against "nan" value for getLinearVelocity)
		--newer happened in solo. Happened once in multi
		local vx = 0; local vy = 0; local vz = 0;
		if self.components[1].node then
			local lvx, lvy, lvz = getLinearVelocity(self.components[1].node);		
			if lvx~=nil and lvx==lvx and lvy~=nil and lvy==lvy and lvz~=nil and lvz==lvz then
				vx, vy, vz = worldDirectionToLocal(self.components[1].node, lvx, lvy, lvz);
			end
		end
		--------------------------------------------------------------
		local movingDirection = 0;
		if vz > 0.001 then
			movingDirection = 1;
		elseif vz < -0.001 then
			movingDirection = -1;
		end;
		local lastSpeedReal = Utils.vector3Length(vx, vy, vz)*0.001;
		local lastMovedDistance = lastSpeedReal*dt;

		self.lastSpeedAcceleration = (lastSpeedReal*movingDirection - self.lastSpeedReal*self.movingDirection)/dt;
		self.lastSpeed = self.lastSpeed*0.5 + lastSpeedReal*0.5;
		self.lastSpeedReal = lastSpeedReal;
		self.movingDirection = movingDirection;
		self.lastMovedDistance = lastMovedDistance;
	else
		local newX, newY, newZ = getWorldTranslation(self.components[1].node);
		if self.lastPosition == nil then
			self.lastPosition = {newX, newY, newZ};
		end;
		local lastMovingDirection = self.movingDirection;
		local dx, dy, dz = worldDirectionToLocal(self.components[1].node, newX-self.lastPosition[1], newY-self.lastPosition[2], newZ-self.lastPosition[3]);
		if dz > 0.001 then
			self.movingDirection = 1;
		elseif dz < -0.001 then
			self.movingDirection = -1;
		else
			self.movingDirection = 0;
		end;
		self.lastMovedDistance = Utils.vector3Length(dx, dy, dz);
		local lastLastSpeedReal = self.lastSpeedReal;
		self.lastSpeedReal = (self.lastMovedDistance/dt);
		self.lastSpeedAcceleration = (self.lastSpeedReal*self.movingDirection - lastLastSpeedReal*lastMovingDirection)/dt;
		self.lastSpeed = self.lastSpeed + ((self.lastSpeedReal-self.lastSpeed) * 0.01);
		self.lastPosition[1], self.lastPosition[2], self.lastPosition[3] = newX, newY, newZ;
	end

	self.updateWheelsTime = self.updateWheelsTime - dt
	if self.isActive then
		self.updateWheelsTime = 3000
	end

	if self.updateWheelsTime > 0 then

		IKUtil.updateIKChains(self.ikChains);

		if self.firstTimeRun then		
			
			WheelsUtil.updateWheelsGraphics(self, dt);

			-- tire tracks and ground sounds
			if self.isEntered and g_currentMission.surfaceSounds ~= nil then
				for _,surfaceSound in pairs(g_currentMission.surfaceSounds) do
					surfaceSound.wheelCount = 0;
				end
			end

			for i,wheel in ipairs(self.wheels) do
				local r, g, b, a, t = nil, nil, nil, nil, nil;
				local surfaceSound = nil;
				local wheelSpeed = 0;
				-- using netinfo couse of tire deformation
				local wx, wy, wz = wheel.netInfo.x, wheel.netInfo.y, wheel.netInfo.z;
				wy = wy - wheel.radius;
				wx = wx + wheel.xOffset;
				wx, wy, wz = localToWorld(wheel.node, wx,wy,wz);

				if self.isServer and self.isAddedToPhysics then
					wheelSpeed = getWheelShapeAxleSpeed(wheel.node, wheel.wheelShape);
					--MR ****************************************
					if self.mrVehicleIsStill then wheelSpeed=0; end
					wheel.mrLastWheelSpeed = wheelSpeed;
					--****************************************************
					local contactObject, contactSubShapeIndex = getWheelShapeContactObject(wheel.node, wheel.wheelShape);
					if contactObject == g_currentMission.terrainRootNode then
						if contactSubShapeIndex <= 0 then
							wheel.contact = Vehicle.WHEEL_GROUND_CONTACT;
						else
							wheel.contact = Vehicle.WHEEL_GROUND_HEIGHT_CONTACT;
						end
					elseif wheel.hasGroundContact and contactObject ~= 0 and getRigidBodyType(contactObject) == "Static" and getUserAttribute(contactObject, "noTireTracks") ~= true then
						wheel.contact = Vehicle.WHEEL_OBJ_CONTACT;
					else
						wheel.contact = Vehicle.WHEEL_NO_CONTACT;
					end;
				else
					if self.movingDirection > 0 then
						wheelSpeed = 1;
					elseif self.movingDirection < 0 then
						wheelSpeed = -1;
					end
				end;

				local isOnField = false;
				if wheel.contact == Vehicle.WHEEL_GROUND_CONTACT then
					local densityBits = getDensityAtWorldPos(g_currentMission.terrainDetailId, wx, wy, wz);
					isOnField = densityBits ~= 0;

					if isOnField then
						r, g, b, a = Utils.getTireTrackColorFromDensityBits(densityBits)
						t = 1
						if self.isEntered and g_currentMission.surfaceNameToSurfaceSound ~= nil then
							surfaceSound = g_currentMission.surfaceNameToSurfaceSound.field;
						end
					else
						r, g, b, a, t = getTerrainAttributesAtWorldPos(g_currentMission.terrainRootNode, wx, wy, wz, true, true, true, true, false);
						if self.isEntered and g_currentMission.surfaceNameToSurfaceSound ~= nil then
							surfaceSound = g_currentMission.materialIdToSurfaceSound[t];
						end
					end

					wheel.dirtAmount = 1;
					wheel.lastColor[1] = r;
					wheel.lastColor[2] = g;
					wheel.lastColor[3] = b;
					wheel.lastColor[4] = a;
					wheel.lastTerrainAttribute = t;
				elseif wheel.contact == Vehicle.WHEEL_OBJ_CONTACT then
					if wheel.dirtAmount > 0 then
						local maxTrackLength = 30 * (1 + g_currentMission.environment.groundWetness);
						local speedFactor = math.min(self:getLastSpeed(), 20) / 20;
						maxTrackLength = maxTrackLength * (2 - speedFactor);
						wheel.dirtAmount = math.max(wheel.dirtAmount - self.lastMovedDistance/maxTrackLength, 0);
						r, g, b = wheel.lastColor[1], wheel.lastColor[2], wheel.lastColor[3];
						a = 0; -- no depth to tyre tracks on road etc.
					end;
					if self.isEntered and g_currentMission.surfaceNameToSurfaceSound ~= nil then
						surfaceSound = g_currentMission.surfaceNameToSurfaceSound.asphalt;
					end
				elseif wheel.contact == Vehicle.WHEEL_GROUND_HEIGHT_CONTACT then
					if self.isServer then
						local movement = self.lastMovedDistance * 1.5; -- move 1.5 per meter
						if wheel.smoothGroundRadius > 0 then
							local heightType = TipUtil.getHeightTypeDescAtWorldPos(wx, wy, wz, wheel.smoothGroundRadius)
							if heightType ~= nil and heightType.allowsSmoothing then
								smoothDensityMapHeightAtWorldPos(TipUtil.terrainDetailHeightUpdater, wx, wy, wz, movement, heightType.index, 0.0, wheel.smoothGroundRadius, wheel.smoothGroundRadius + 1);
								if Vehicle.debugRendering then
									DebugUtil.drawDebugCircle(wx,wy,wz, wheel.smoothGroundRadius, 10);
								end
							end
						end
						if wheel.additionalWheels ~= nil then
							for _, additionalWheel in pairs(wheel.additionalWheels) do
								local refNode = wheel.node;
								if wheel.repr ~= wheel.driveNode then
									refNode = wheel.repr;
								end
								local xShift,yShift,zShift = localToLocal(additionalWheel.wheelTire, refNode, additionalWheel.xOffset,0,0);
								local wx,wy,wz = localToWorld(refNode, xShift, yShift-additionalWheel.radius, zShift);
								local heightType = TipUtil.getHeightTypeDescAtWorldPos(wx, wy, wz, additionalWheel.smoothGroundRadius)
								if heightType ~= nil and heightType.allowsSmoothing then
									smoothDensityMapHeightAtWorldPos(TipUtil.terrainDetailHeightUpdater, wx, wy, wz, movement, heightType.index, 0.0, additionalWheel.smoothGroundRadius, additionalWheel.smoothGroundRadius + 1);
									if Vehicle.debugRendering then
										DebugUtil.drawDebugCircle(wx,wy,wz, additionalWheel.smoothGroundRadius, 10);
									end
								end
							end
						end
					end
				end;
				
				wheel.mrLastA = a;
				wheel.mrIsOnField = isOnField;

				--[[ directly called by "mrManageWheelsFriction"
				--MR modification ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------						
				if self.isServer and not wheel.mrNotAWheel then					
					if wheel.contact ~= Vehicle.WHEEL_NO_CONTACT then
						local groundType = WheelsUtil.getGroundType(isOnField, wheel.contact ~= Vehicle.WHEEL_GROUND_CONTACT, a);
						--local coeff = WheelsUtil.getTireFriction(wheel.tireType, groundType, g_currentMission.environment.groundWetness);
						local coeff = WheelsUtil.mrGetTireFriction(wheel, groundType, g_currentMission.environment.groundWetness);
						--if coeff ~= wheel.tireGroundFrictionCoeff then
						wheel.tireGroundFrictionCoeff = coeff;						
						--end
					end					
					self:updateWheelTireFriction(wheel);
					------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
				end--]]

				if wheel.tireTrackIndex ~= nil then
					if r ~= nil then
						local ux,uy,uz = localDirectionToWorld(self.rootNode, 0,1,0);
						-- we are using dirtAmount as alpha value -> realistic dirt fadeout

						g_currentMission.tireTrackSystem:addTrackPoint(wheel.tireTrackIndex, wx, wy, wz, ux, uy, uz, r, g, b, wheel.dirtAmount, a, wheelSpeed);
						if wheel.additionalWheels ~= nil then
							for _, additionalWheel in pairs(wheel.additionalWheels) do
								if additionalWheel.tireTrackIndex ~= nil then
									local wx, wy, wz = worldToLocal(wheel.node, getWorldTranslation(additionalWheel.wheelTire));
									wy = wy - wheel.radius;
									wx = wx + wheel.xOffset;
									wx, wy, wz = localToWorld(wheel.node, wx,wy,wz);
									g_currentMission.tireTrackSystem:addTrackPoint(additionalWheel.tireTrackIndex, wx, wy, wz, ux, uy, uz, r, g, b, wheel.dirtAmount, a, wheelSpeed);
								end
							end;
						end;
					else
						g_currentMission.tireTrackSystem:cutTrack(wheel.tireTrackIndex);
						if wheel.additionalWheels ~= nil then
							for _, additionalWheel in pairs(wheel.additionalWheels) do
								if additionalWheel.tireTrackIndex ~= nil then
									g_currentMission.tireTrackSystem:cutTrack(additionalWheel.tireTrackIndex);
								end
							end;
						end;
					end;
				end


				if surfaceSound ~= nil then
					surfaceSound.wheelCount = surfaceSound.wheelCount + 1;
				end

			end; --end for each wheels
			
			
			-- MR ---------------------------------------------------------------------------------------------------------------------------------------------------
			if self.isServer then
				self:mrManageWheelsFriction(dt);
			end;
			-- END MR ---------------------------------------------------------------------------------------------------------------------------------------------------
					
			
			-- MR ---------------------------------------------------------------------------------------------------------------------------------------------------
			--measure tractive efficiency = drawbar power / axle power
			--Removed since "computeWheelShapeTireForce is not reliable at all (it seems the actual longitudinal force depends on longForce and latForce !)
			--[[if Vehicle.debugRendering and self.motor~=nil and self.isControlled and self:getIsActiveForInput() then
				local sumLongForces=0;				
				for i,wheel in ipairs(self.wheels) do
					
					local longForce, latForce = computeWheelShapeTireForces(wheel.node, wheel.wheelShape, wheel.mrLastLongSlip, wheel.mrLastLatSlip, wheel.mrLastTireLoad);
					sumLongForces = sumLongForces + longForce;		
					
					--if i==3 then
					--	Vehicle.testWheelFrictionForce(self, wheel);
					--end;
				end
				local tractiveEfficiency = 0;
				if self.motor.mrLastTransmissionTorque*self.motor.clutchRpm>0 and self.lastSpeedReal*1000>0.1 then		
					if self.mrDebugLastTotalResistance==nil then self.mrDebugLastTotalResistance=0; end
					local drawbarPower = self.lastSpeedReal*1000*(sumLongForces-self.mrDebugLastTotalResistance);
					local axlePower = self.motor.mrLastTransmissionTorque*self.motor.clutchRpm/9.5488;
					if self.mrDebugLastTractiveEfficiency==nil then self.mrDebugLastTractiveEfficiency=0.5; end
					if axlePower>0 then
						tractiveEfficiency = 0.98 * self.mrDebugLastTractiveEfficiency + 0.02 * drawbarPower/axlePower;
					else
						tractiveEfficiency = 0.98 * self.mrDebugLastTractiveEfficiency;
					end
					self.mrDebugLastTractiveEfficiency = tractiveEfficiency;
					
					--print(" axlePower="..tostring(axlePower) .. " - drawbarPower="..tostring(drawbarPower) .. " - lastTorqueToTransmission="..tostring(self.motor.mrLastTransmissionTorque));
					--print("     drawbarPower - speed="..tostring(self.lastSpeedReal*1000*3.6).." - sumLongForces="..tostring(sumLongForces).." - self.mrDebugLastTotalResistance="..tostring(self.mrDebugLastTotalResistance));
				end;
				renderText(0.2, 0.59, getCorrectTextSize(0.02), string.format("tractive Efficiency: %1.2f", tractiveEfficiency));				
			end;--]]
			-----------------------------------------------------------------------------------------------------------------------------------------------------
			
			
			-- adjust volume of ground sounds, only in entered steerable vehicle
			if self.isEntered and g_currentMission.surfaceSounds ~= nil then

				local wheelCount = table.getn(self.wheels);
				local noSound = math.abs(self:getLastSpeed()) < 1;

				local debugString = "";

				for _,surfaceSound in pairs(g_currentMission.surfaceSounds) do
					if surfaceSound.sample.sample ~= nil then

						local currentVolume = surfaceSound.currentVolume;
						local targetVolume = surfaceSound.sample.volume * (surfaceSound.wheelCount/wheelCount);

						if currentVolume < targetVolume then
							currentVolume = math.min(targetVolume, currentVolume + surfaceSound.sample.volume * dt/125);
						elseif currentVolume > targetVolume then
							currentVolume = math.max(targetVolume, currentVolume - surfaceSound.sample.volume * dt/125);
						end

						if noSound then
							currentVolume = 0;
						end

						if currentVolume > 0 then
							if not SoundUtil.isSamplePlaying(surfaceSound.sample) then
								SoundUtil.playSample(surfaceSound.sample, 0, 0, 0);
							end
							if surfaceSound.pitchScale ~= nil then
								local pitch = surfaceSound.sample.pitchOffset + self:getLastSpeed()*surfaceSound.pitchScale;
								SoundUtil.setSamplePitch(surfaceSound.sample, pitch);
							end
						elseif SoundUtil.isSamplePlaying(surfaceSound.sample) then
							SoundUtil.stopSample(surfaceSound.sample);
						end

						if currentVolume ~= surfaceSound.currentVolume then
							SoundUtil.setSampleVolume(surfaceSound.sample, currentVolume);
						end

						surfaceSound.currentVolume = currentVolume;

						debugString = debugString .. string.format("%s=%.1f%%  ", tostring(surfaceSound.name), 100*currentVolume/surfaceSound.sample.volume);
					end
				end
				--print(debugString);
				if self.isActive and Vehicle.debugRendering then
					renderText(0.2, 0.02, getCorrectTextSize(0.02), "volumes: " .. debugString);
				end

			end


			if Vehicle.debugMoveAttacherJoints then
				if self.selectedImplement ~= nil then
					local jointDescIndex = self.selectedImplement.jointDescIndex;
					local jointDesc = self.attacherJoints[jointDescIndex];
					if jointDesc.rotationNode ~= nil then
						local moveUpper, _ = InputBinding.getInputAxis(InputBinding.AXIS_FRONTLOADER_ARM);
						local moveLower, _ = InputBinding.getInputAxis(InputBinding.AXIS_FRONTLOADER_TOOL);
						if moveUpper ~= 0 then
							jointDesc.upperRotation[1] = jointDesc.upperRotation[1] + math.rad(moveUpper*(2/1000)*dt);
							jointDesc.moveAlpha = jointDesc.moveAlpha - 0.001;
							print("upperRotation: " ..math.deg(jointDesc.upperRotation[1]));
						end;
						if moveLower ~= 0 then
							jointDesc.lowerRotation[1] = jointDesc.lowerRotation[1] + math.rad(moveLower*(2/1000)*dt);
							jointDesc.moveAlpha = jointDesc.moveAlpha - 0.001;
							print("lowerRotation: " ..math.deg(jointDesc.lowerRotation[1]));
						end;
					end;
				end
			end

		end;
		if self.showDetachingNotAllowedTime > 0 then
			self.showDetachingNotAllowedTime = self.showDetachingNotAllowedTime - dt;
		end;
	else
		self.showDetachingNotAllowedTime = 0;
		if g_currentMission.tireTrackSystem ~= nil then
			for _,wheel in ipairs(self.wheels) do
				if wheel.tireTrackIndex ~= nil then
					g_currentMission.tireTrackSystem:cutTrack(wheel.tireTrackIndex);
				end
				if wheel.additionalWheels ~= nil then
					for _, additionalWheel in pairs(wheel.additionalWheels) do
						if additionalWheel.tireTrackIndex ~= nil then
							g_currentMission.tireTrackSystem:cutTrack(additionalWheel.tireTrackIndex);
						end
					end;
				end;
			end;
		end
	end;

	for k,v in pairs(self.specializations) do
		v.update(self, dt);
	end;

	for _,v in ipairs(self.specializations) do
		if v.postUpdate ~= nil then
			v.postUpdate(self, dt);
		end;
	end;

	if Vehicle.debugAttributeRendering then
		Vehicle.drawDebugAttributeRendering(self);		
	end;

	self.firstTimeRun = true;
end;
Vehicle.update = Utils.overwrittenFunction(Vehicle.update, Vehicle.mrUpdate);



--************************************************************************************************************************************************************
-- add rolling resistance and air resistance (drag)
Vehicle.mrUpdate2 = function(self, dt)
		
	if self.isServer and self.mrIsMrVehicle then
		--if not self.hasStopped and self.firstTimeRun and self.movingDirection ~= 0 and self.isActive then 
		
		--if self.mrVehicleIsStill then
		--	print(tostring(g_currentMission.time) .. " - vehicle is still = " .. self.configFileName);
		--end
		
		
		if not self.mrVehicleIsStill then
			
			local spd = self.lastSpeedReal*1000; -- meter per second			
			local avgSlip = 0;			
			local totalResistanceForce = 0;
			local airResistance = 0;
			local totalSlippingResistance = 0;
			local totalWheelRollingResistance = 0;				
			
			if spd>0.1 then --0.1 = 0.36kph
				--local totalWheelSlip = 0;
				--local totalWheelSlipNumWheels = 0;							
				for _,wheel in ipairs(self.wheels) do
					if wheel.hasGroundContact and not wheel.mrNotAWheel then
						--add rolling resistance function of load on the tire					
						local wRes = wheel.mrLastTireLoad * wheel.mrCurrentRollingResistance;
						--print("test - mrLastTireLoad = "..tostring(wheel.mrLastTireLoad) .. " - mrCurrentRollingResistance="..tostring(wheel.mrCurrentRollingResistance) .. " - width="..tostring(wheel.mrTotalContactWidth));
						totalWheelRollingResistance = totalWheelRollingResistance + wRes;
						--only for "motorized" vehicles
						if self.motorizedNode~=nil then							
							--slipping resistance when wheel "dig" into the ground
							if wheel.mrRealLongSlip>0.1 and wheel.mrLastTireLoad>0 then --only if slip>10%
							--if wheel.mrSlippingResistanceEnabled and wheel.mrLastTireLoad>0 then
								totalSlippingResistance = totalSlippingResistance + wRes * 4 * (wheel.mrRealLongSlip-0.1) * math.max(0, 1 -20*wheel.mrTotalContactWidth*wheel.radius/wheel.mrLastTireLoad);--factor function of wheel width and radius (the wider the wheel, the smaller the slipping resistance. The taller the wheel, the smaller the slipping resistance)
								--totalSlippingResistance = totalSlippingResistance + wRes * 4 * wheel.mrRealLongSlip * math.max(0, 1 -20*wheel.mrTotalContactWidth*wheel.radius/wheel.mrLastTireLoad);
							end
							--totalWheelSlip = totalWheelSlip + wheel.mrLastLongSlip;
							--totalWheelSlipNumWheels = totalWheelSlipNumWheels + 1;
						end;						
					end;
				end;
				--if totalWheelSlipNumWheels>0 then
				--	avgSlip = totalWheelSlip/totalWheelSlipNumWheels;				
				--end				
			end;			
			--self.mrAvgWheelSlip = self.mrAvgWheelSlip*0.95 + avgSlip*0.05;
			
			--add air resistance only if not attached to another vehicle
			if spd>4 and not self.attacherVehicle then				
				-- SCx = S multiply Cx where S= 2d surface of the vehicle (front view) and Cx = drag coefficient (1 is a good approximation for a farming vehicle)
				-- good approximation for farming vehicle = 0.85 x width x height
				-- example : tractor 2.5m width and 3m high => 0.85 x 2.5 x 3 = 6.375
				--english term = "drag area"				
				airResistance = 0.00061*RealisticUtils.getDragArea(self)*spd^2; -- 0.5 * uAir * dragArea * spd^2  (uAir = 1.22)
			end
			
			--smooth wheelRollingResistance	
			--20170317 - fix bug : self.mrRollingResistanceS was never 0, and so, a force (even minimal) was always apply to the vehicle = vehicle never still
			if totalWheelRollingResistance==0 then
				self.mrRollingResistanceS = 0;
			else
				self.mrRollingResistanceS = 0.99 * self.mrRollingResistanceS + 0.01 * totalWheelRollingResistance;
			end
			
			local totalResistanceForce = airResistance + self.mrRollingResistanceS + totalSlippingResistance;
						
			if Vehicle.debugRendering and not self.attacherVehicle and spd>0.1 then
				if RealisticUtils.mrDebugValue3 then
					totalResistanceForce = tonumber(RealisticUtils.mrDebugValue3);
				end
			end;			
			
			if totalResistanceForce>0 then
				local px,py,pz = getCenterOfMass(self.components[1].node);
				if spd>0.3 then --0.3 = 1.08
					local vx, vy, vz = getLinearVelocity(self.components[1].node);				
					if vx==nil or vy==nil or vz==nil then				
						RealisticUtils.printWarning("Vehicle.mrUpdate2", "getLinearVelocity has returned a nil value. Check " .. tostring(self.configFileName), true);				
					else
										
						local spd = Utils.vector3Length(vx, vy, vz);						
						addForce(self.components[1].node, -totalResistanceForce*vx/spd,-totalResistanceForce*vy/spd,-totalResistanceForce*vz/spd, px,py,pz, true);					
					end
				else
					local dx,dy,dz = localDirectionToWorld(self.components[1].node, 0, 0, -totalResistanceForce*self.movingDirection);					
					addForce(self.components[1].node, dx,dy,dz, px,py,pz, true);	
				end
			end;
			
			if Vehicle.debugRendering and self:getIsActiveForInput() then	
				renderText(0.35, 0.55, getCorrectTextSize(0.02), string.format("Rolling resistance force: %1.2f", self.mrRollingResistanceS));
				renderText(0.35, 0.53, getCorrectTextSize(0.02), string.format("Air resistance force: %1.2f", airResistance));
				renderText(0.35, 0.51, getCorrectTextSize(0.02), string.format("Slipping Resistance force: %1.2f", totalSlippingResistance));
				self.mrDebugLastTotalResistance = totalResistanceForce;
			end
			
		end;
		
		if Vehicle.debugRendering and self.isEntered and self.isClient and self.isControlled then		
		
			if self.selectedImplement ~= nil and self.selectedImplement.object ~= nil then
				--MR : display total mass on the wheels
				Vehicle.drawDebugMassOnTheWheels(self.selectedImplement.object);
			else
				Vehicle.drawDebugMassOnTheWheels(self);
			end;
			
			if self.mrDebugLastAccPedal==nil then self.mrDebugLastAccPedal = 0; end;
			renderText(0.70, 0.50, getCorrectTextSize(0.02), string.format("Acc: %1.2f", self.mrDebugLastAccPedal));
			
			if self:getIsActiveForInput() then	
				
				--measure stopping distance		
				--init variables
				if self.mrDebugMovedDistance == nil then self.mrDebugMovedDistance = 0;	end;
				if self.mrDebugReadyForStoppingDistance == nil then self.mrDebugReadyForStoppingDistance = true; end;
				
				local distance = 0;
				if self.mrDebugReadyForStoppingDistance==false and self.lastSpeedReal*3600<0.1 then
					--vehicle stopped, compute distance and reset data
					self.mrDebugMovedDistance = self.mrDebugMovedDistance + self.lastMovedDistance;			
					self.mrDebugReadyForStoppingDistance = true;
				elseif self.mrDebugReadyForStoppingDistance then
					if self.axisForward>=0 and self.cruiseControl.state ~= Drivable.CRUISECONTROL_STATE_ACTIVE and self.lastSpeedReal*3600>1 then
						--start measuring
						self.mrDebugReadyForStoppingDistance = false;
						self.mrDebugMovedDistance = 0;
					end;
				else
					--already measuring distance
					self.mrDebugMovedDistance = self.mrDebugMovedDistance + self.lastMovedDistance;
				end;
				renderText(0.34, 0.61, getCorrectTextSize(0.02), string.format("Braking distance = %.1fm", self.mrDebugMovedDistance));
				
				
				--measure 0-100kph and 0-40kph times
				--init variables
				if self.DebugTimeStartTime == nil then self.DebugTimeStartTime = 0;	end;				
				
				if math.abs(self.axisForward)<0.01 and self.lastSpeedReal*3600<0.1 then
					--ready to measure time
					self.DebugTimeStartTime = g_currentMission.time;
					self.mrDebug40Time = nil;
					self.mrDebug100Time = nil;
				else					
					if self.mrDebug40Time == nil and self.lastSpeedReal*3600>=40 then
						self.mrDebug40Time = g_currentMission.time-self.DebugTimeStartTime;
					elseif self.mrDebug100Time == nil and self.lastSpeedReal*3600>=100 then
						self.mrDebug100Time = g_currentMission.time-self.DebugTimeStartTime;
					end
				end
				renderText(0.34, 0.63, getCorrectTextSize(0.02), string.format("Acc times 0 to 40/100kph = %2.2f/%2.2f", Utils.getNoNil(self.mrDebug40Time, 0)/1000,Utils.getNoNil(self.mrDebug100Time, 0)/1000));
				
				
				--show the current environment ground wetness
				renderText(0.2, 0.63, getCorrectTextSize(0.02), string.format("Ground wetness: %1.2f", g_currentMission.environment.groundWetness));
				
				--show if the brake are applyied
				renderText(0.2, 0.61, getCorrectTextSize(0.02), string.format("Brake: %s", tostring(self.mrIsBraking)));

				--display the average wheel slip
				renderText(0.2, 0.57, getCorrectTextSize(0.02), string.format("Avg Slip: %3.1f%%", 100*self.mrAvgWheelSlip));
				
				--display the last engine output power delivered
				local lastEngineOutputPower = self.motor.equalizedMotorRpm*math.pi/30*self.motor.mrLastEngineOutputTorque; -- KW
				
				renderText(0.2, 0.55, getCorrectTextSize(0.02), string.format("Engine Output Power\n(equivalent PTO power) :\n%3.0fKW/%3.0fHp", lastEngineOutputPower, lastEngineOutputPower/0.7355));
				
				
				
				if self.motor then
					renderText(0.135, 0.585, getCorrectTextSize(0.02), string.format("MR: %3.1f", self.motor.mrLastDummyGearRatio));
				end
				
			end;
			
		end;

	end--end self.isServer
	
end;
Vehicle.update = Utils.appendedFunction(Vehicle.update, Vehicle.mrUpdate2);


--************************************************************************************************************************************************************
--add support for "$mr" tag when loading a wheel xml file
Vehicle.mrLoadWheelDataFromXML = function(self, wheel, xmlFilename, wheelConfigIndex, isLeft, isRealWheel, xRotOffset)
	
	--MR MODIFICATION
	--check if we want to load the wheel xml file from the "moreRealistic" folder	
	xmlFilename = RealisticUtils.getFilename(xmlFilename, self.baseDirectory);
	--END MR MODIFICATION
	----------------------------------------------------------------------------------------	
	local xmlFile = loadXMLFile("TempConfig", xmlFilename);	

	if xmlFile ~= nil then
		local i = 0;
		local wheelConfigFound = false
		while true do
			local configKey = string.format("wheel.configurations.configuration(%d)", i)
			if not hasXMLProperty(xmlFile, configKey) then
				break;
			end;
			if getXMLString(xmlFile, configKey.."#index") == wheelConfigIndex then
				wheelConfigFound = true

				local key = "indexLeft"
				if not isLeft then
					key = "indexRight"
				end
				wheel.isLeft = isLeft

				wheel.tireFilename = getXMLString(xmlFile, configKey..".tire#filename")
				wheel.tireIndex = Utils.getNoNil(getXMLString(xmlFile, configKey..".tire#"..key), "0|0")
				wheel.outerRimFilename = getXMLString(xmlFile, configKey..".outerRim#filename")
				wheel.outerRimIndex = Utils.getNoNil(getXMLString(xmlFile, configKey..".outerRim#index"), "0|0")
				wheel.outerRimWidthAndDiam = Utils.getVectorNFromString(getXMLString(xmlFile, configKey..".outerRim#widthAndDiam"), 2)
				wheel.outerRimScale = Utils.getVectorNFromString(Utils.getNoNil(getXMLString(xmlFile, configKey..".outerRim#scale"), "1 1 1"), 3)
				wheel.innerRimFilename = getXMLString(xmlFile, configKey..".innerRim#filename")
				wheel.innerRimIndex = Utils.getNoNil(getXMLString(xmlFile, configKey..".innerRim#"..key), "0|0")
				wheel.innerRimWidthAndDiam = Utils.getVectorNFromString(getXMLString(xmlFile, configKey..".innerRim#widthAndDiam"), 2)
				wheel.innerRimOffset = Utils.getNoNil(getXMLFloat(xmlFile, configKey..".innerRim#offset"), 0)
				wheel.innerRimScale = Utils.getVectorNFromString(Utils.getNoNil(getXMLString(xmlFile, configKey..".innerRim#scale"), "1 1 1"), 3)
				wheel.hubFilename = getXMLString(xmlFile, configKey..".hub#filename")
				wheel.hubIndex = Utils.getNoNil(getXMLString(xmlFile, configKey..".hub#"..key), "0|0")
				wheel.hubOffset = Utils.getNoNil(getXMLFloat(xmlFile, configKey..".hub#offset"), 0)
				wheel.hubScale = Utils.getVectorNFromString(Utils.getNoNil(getXMLString(xmlFile, configKey..".hub#scale"), "1 1 1"), 3)
				wheel.hubWidthAndDiam = Utils.getVectorNFromString(getXMLString(xmlFile, configKey..".hub#widthAndDiam"), 2)
				wheel.additionalFilename = getXMLString(xmlFile, configKey..".additional#filename")
				wheel.additionalIndex = Utils.getNoNil(getXMLString(xmlFile, configKey..".additional#"..key), "0|0")
				wheel.additionalOffset = Utils.getNoNil(getXMLFloat(xmlFile, configKey..".additional#offset"), 0)
				wheel.additionalScale = Utils.getVectorNFromString(Utils.getNoNil(getXMLString(xmlFile, configKey..".additional#scale"), "1 1 1"), 3)
				wheel.additionalMass = Utils.getNoNil(getXMLFloat(xmlFile, configKey..".additional#mass"), 0)
				wheel.additionalWidthAndDiam = Utils.getVectorNFromString(getXMLString(xmlFile, configKey..".additional#widthAndDiam"), 2)
				break;
			end
			i = i + 1;
		end

		if not wheelConfigFound then
			print("Error: wheelConfigIndex '" .. wheelConfigIndex .. "' not found in '" .. xmlFilename .. "'")
			return
		end

		wheel.radius = Utils.getNoNil(getXMLFloat(xmlFile, "wheel.size#radius"), 0.5);
		wheel.width = Utils.getNoNil(getXMLFloat(xmlFile, "wheel.size#width"), wheel.radius * 0.8);
		wheel.xOffset = Utils.getNoNil(getXMLFloat(xmlFile, "wheel.size#xOffset"), 0);
		wheel.maxDeformation = Utils.getNoNil(getXMLFloat(xmlFile, "wheel.size#maxDeformation"), 0);

		wheel.mass = Utils.getNoNil(getXMLFloat(xmlFile, "wheel.physics#mass"), wheel.mass) + wheel.additionalMass
		wheel.maxLongStiffness = Utils.getNoNil(getXMLFloat(xmlFile, "wheel.physics#maxLongStiffness"), wheel.maxLongStiffness); -- [t / rad]
		wheel.maxLatStiffness = Utils.getNoNil(getXMLFloat(xmlFile, "wheel.physics#maxLatStiffness"), wheel.maxLatStiffness); -- xml is ratio to restLoad [1/rad], final value is [t / rad]

		wheel.smoothGroundRadius = Utils.getNoNil(getXMLFloat(xmlFile, "wheel.physics#smoothGroundRadius"), math.max(0.6, wheel.width*0.75));

		if isRealWheel then
			wheel.frictionScale = Utils.getNoNil(getXMLFloat(xmlFile, "wheel.physics#frictionScale"), wheel.frictionScale);
			local tireTypeName = getXMLString(xmlFile, "wheel.tire#type");
			if tireTypeName ~= nil then
				local tireType = WheelsUtil.getTireType(tireTypeName);
				if tireType ~= nil then
					wheel.tireType = tireType;
				else
					print("Warning: Failed to find tire type '"..tireTypeName.."'.");
				end
			end;

			wheel.maxLatStiffnessLoad = Utils.getNoNil(getXMLFloat(xmlFile, "wheel.physics#maxLatStiffnessLoad"), wheel.maxLatStiffnessLoad); -- xml is ratio to restLoad, final value is [t]
		end;

		wheel.tireTrackAtlasIndex = Utils.getNoNil(getXMLInt(xmlFile, "wheel.tire#atlasIndex"), 0);

		local loadWheelPart = function(wheel, parent, name, filename, index, offset, widthAndDiam, scale)
			if filename == nil then
				return
			end
			local i3dNode = Utils.loadSharedI3DFile(filename, self.baseDirectory, false, false, false);
			if i3dNode ~= 0 then
				wheel[name] = Utils.indexToObject(i3dNode, index);
				link(parent, wheel[name]);
				delete(i3dNode);

				if offset ~= 0 then
					local dir = 1
					if not wheel.isLeft then
						dir = -1
					end
					setTranslation(wheel[name], offset*dir, 0, 0)
				end
				if scale ~= nil then
					setScale(wheel[name], scale[1], scale[2], scale[3])
				end
				if widthAndDiam ~= nil then
					if getHasShaderParameter(wheel[name], "widthAndDiam") then
						setShaderParameter(wheel[name], "widthAndDiam", widthAndDiam[1], widthAndDiam[2], 0, 0, false)
					end
				end
			end;
		end

		loadWheelPart(wheel, wheel.linkNode,  "wheelTire",       wheel.tireFilename,       wheel.tireIndex,       0,                      nil,                          nil)
		loadWheelPart(wheel, wheel.wheelTire, "wheelOuterRim",   wheel.outerRimFilename,   wheel.outerRimIndex,   0,                      wheel.outerRimWidthAndDiam,   wheel.outerRimScale)
		loadWheelPart(wheel, wheel.wheelTire, "wheelInnerRim",   wheel.innerRimFilename,   wheel.innerRimIndex,   wheel.innerRimOffset,   wheel.innerRimWidthAndDiam,   wheel.innerRimScale)
		loadWheelPart(wheel, wheel.wheelTire, "wheelHub",        wheel.hubFilename,        wheel.hubIndex,        wheel.hubOffset,        wheel.hubWidthAndDiam,        wheel.hubScale)
		loadWheelPart(wheel, wheel.wheelTire, "wheelAdditional", wheel.additionalFilename, wheel.additionalIndex, wheel.additionalOffset, wheel.additionalWidthAndDiam, wheel.additionalScale)

		setRotation(wheel.wheelTire, xRotOffset, 0, 0)

		local x, y, z, _ = getShaderParameter(wheel.wheelTire, "morphPosition");
		setShaderParameter(wheel.wheelTire, "morphPosition", x, y, z, 0, false);

		local color = Utils.getNoNil(Utils.getNoNil(wheel.color, Vehicle.getColorByConfigId(self, "rimColor", self.configurations["rimColor"])), self.rimColor);
		if color ~= nil then
			local r,g,b,a = unpack(color);
			if wheel.wheelOuterRim ~= nil then
				setShaderParameter(wheel.wheelOuterRim, "colorScale", r, g, b, a, false);
			end
			if wheel.wheelInnerRim ~= nil then
				setShaderParameter(wheel.wheelInnerRim, "colorScale", r, g, b, a, false);
			end
		end

		local additionalColor = Utils.getNoNil(wheel.additionalColor, color);
		if wheel.wheelAdditional ~= nil and additionalColor ~= nil then
			local r,g,b,a = unpack(additionalColor);
			setShaderParameter(wheel.wheelAdditional, "colorScale", r, g, b, a, false);
		end

		local wheelAxisColor = Utils.getNoNil(wheel.axisColor, self.axisColor)
		if wheelAxisColor ~= nil then
			if wheel.wheelHub ~= nil then
				local r,g,b,a = unpack(wheelAxisColor);
				setShaderParameter(wheel.wheelHub, "colorScale", r, g, b, a, false);
			end
		end

		delete(xmlFile);
	end;
	
end;

--allow us to override position of the wheelShape directly in the xml without having to modify the i3d file
--especially useful for adding "dummy" wheels to help reducing friction forces between the implement and the ground (this is far better to use the powerConsumer specialization for this purpose)
Vehicle.mrLoadDynamicWheelDataFromXML2 = function(self, xmlFile, key, wheelnamei, wheel)

	--override wheel position if specified in the xml
	wheel.positionX = Vehicle.getConfigurationValue(xmlFile, key, wheelnamei, "#mrPositionX", getXMLFloat, wheel.positionX, nil, fallbackOldKey);	
	wheel.positionY = Vehicle.getConfigurationValue(xmlFile, key, wheelnamei, "#mrPositionY", getXMLFloat, wheel.positionY, nil, fallbackOldKey);	
	wheel.positionZ = Vehicle.getConfigurationValue(xmlFile, key, wheelnamei, "#mrPositionZ", getXMLFloat, wheel.positionZ, nil, fallbackOldKey);	
	
	wheel.startPositionX = Vehicle.getConfigurationValue(xmlFile, key, wheelnamei, "#mrPositionX", getXMLFloat, wheel.startPositionX, nil, fallbackOldKey);	
	wheel.startPositionY = Vehicle.getConfigurationValue(xmlFile, key, wheelnamei, "#mrPositionY", getXMLFloat, wheel.startPositionY, nil, fallbackOldKey);	
	wheel.startPositionZ = Vehicle.getConfigurationValue(xmlFile, key, wheelnamei, "#mrPositionZ", getXMLFloat, wheel.startPositionZ, nil, fallbackOldKey);	

	wheel.mrIsDummyWheel = Vehicle.getConfigurationValue(xmlFile, key, wheelnamei, "#mrDummyWheel", getXMLBool, false, nil, fallbackOldKey);
	
	if wheel.mrIsDummyWheel then
		wheel.createTipOcclusionArea = false;
		wheel.isSynchronized = false;
		wheel.mass = 0.01;		
		wheel.mrDummyNode = createTransformGroup("dummyWheelNode");
		--link(self.components[1].node, wheel.mrDummyNode);
		link(wheel.driveNode, wheel.mrDummyNode);
		wheel.repr = wheel.mrDummyNode;
		wheel.driveNode = wheel.mrDummyNode;
		wheel.linkNode = wheel.mrDummyNode;		
	end
	
end
Vehicle.loadDynamicWheelDataFromXML = Utils.prependedFunction(Vehicle.loadDynamicWheelDataFromXML, Vehicle.mrLoadDynamicWheelDataFromXML2);

--************************************************************************************************************************************************************
--FIX bug in Vehicle.loadDynamicWheelDataFromXML line 906 => the tireType specified in the wheel xml file is not taken into account when there is no tiretype defined in the vehicle xml file
Vehicle.mrLoadDynamicWheelDataFromXML = function(self, xmlFile, key, wheelnamei, wheel)

	local fallbackOldKey = "vehicle.wheels";	
	
	-- 1. check if there is a tireType defined in the vehicle xml file
	local tireTypeName = Vehicle.getConfigurationValue(xmlFile, key, wheelnamei, "#tireType", getXMLString, "notDefined", nil, fallbackOldKey);
	
	if tireTypeName=="notDefined" then	
		-- 2. check if there is a wheel xml file
		local wheelXmlFilename = Vehicle.getConfigurationValue(xmlFile, key, wheelnamei, "#filename", getXMLString, nil, nil, fallbackOldKey);		
		if wheelXmlFilename ~= nil and wheelXmlFilename ~= "" then 
			-- 3. check if there is a tireType defined in the wheel xml file
			local wheelConfigIndex = Vehicle.getConfigurationValue(xmlFile, key, wheelnamei, "#configIndex", getXMLString, "basic", nil, fallbackOldKey);
			local xmlFilename = RealisticUtils.getFilename(wheelXmlFilename, self.baseDirectory);
			local xmlFileW = loadXMLFile("TempConfig", xmlFilename);
			if xmlFileW ~= nil then			
				local tireTypeNameW = getXMLString(xmlFileW, "wheel.tire#type");				
				if tireTypeNameW ~= nil then
					local tireType = WheelsUtil.getTireType(tireTypeNameW);					
					if tireType ~= nil then
						-- 4. replace the "default" mud type by the one specified in the wheel xml file if present and valid
						--print("test - wheel.tire typename="..tostring(tireTypeNameW) .. " - old int=" .. tostring(wheel.tireType) .. " - new int="..tostring(tireType));
						wheel.tireType = tireType;						
					end;
				end;
			end;
			delete(xmlFileW);
		end;
	end;
	
	--load default damper settings if the vehicle is mr
	if self.mrIsMrVehicle then
		local damper = Vehicle.getConfigurationValue(xmlFile, key, wheelnamei, "#damper", getXMLFloat, 0, nil, fallbackOldKey);			
		wheel.damperCompressionLowSpeed = Vehicle.getConfigurationValue(xmlFile, key, wheelnamei, "#damperCompressionLowSpeed", getXMLFloat, damper, nil, fallbackOldKey);	
		wheel.damperRelaxationLowSpeed = Vehicle.getConfigurationValue(xmlFile, key, wheelnamei, "#damperRelaxationLowSpeed", getXMLFloat, damper, nil, fallbackOldKey);	
		wheel.damperCompressionHighSpeed = Vehicle.getConfigurationValue(xmlFile, key, wheelnamei, "#damperCompressionHighSpeed", getXMLFloat, damper, nil, fallbackOldKey);	
		wheel.damperRelaxationHighSpeed = Vehicle.getConfigurationValue(xmlFile, key, wheelnamei, "#damperRelaxationHighSpeed", getXMLFloat, damper, nil, fallbackOldKey);		
	end
	
	
	--load mr wheel parameters from vehicle xml file
	wheel.mrLateralStiffnessScale = Vehicle.getConfigurationValue(xmlFile, key, wheelnamei, "#mrLateralStiffnessScale", getXMLFloat, 1, nil, fallbackOldKey);
	wheel.mrLongitudinalStiffnessScale = Vehicle.getConfigurationValue(xmlFile, key, wheelnamei, "#mrLongitudinalStiffnessScale", getXMLFloat, 1, nil, fallbackOldKey);	
	wheel.mrNotAWheel = Vehicle.getConfigurationValue(xmlFile, key, wheelnamei, "#mrNotAWheel", getXMLBool, false or wheel.mrIsDummyWheel, nil, fallbackOldKey);	
	wheel.mrPressureFx = Vehicle.getConfigurationValue(xmlFile, key, wheelnamei, "#mrPressureFx", getXMLFloat, 1, nil, fallbackOldKey);	
	
	wheel.mrForcePointOffsetX = Vehicle.getConfigurationValue(xmlFile, key, wheelnamei, "#mrForcePointOffsetX", getXMLFloat, 0.5*wheel.width, nil, fallbackOldKey);	
	wheel.mrForcePointOffsetY = Vehicle.getConfigurationValue(xmlFile, key, wheelnamei, "#mrForcePointOffsetY", getXMLFloat, 0.5*wheel.radius, nil, fallbackOldKey);

	wheel.mrStabilizingSteeringAngle = Vehicle.getConfigurationValue(xmlFile, key, wheelnamei, "#mrStabilizingSteeringAngle", getXMLFloat, 0, nil, fallbackOldKey); --degree
	--degree to rad
	wheel.mrStabilizingSteeringAngle = wheel.mrStabilizingSteeringAngle * math.pi/180;
	if wheel.positionX<0 then
		wheel.mrStabilizingSteeringAngle = -wheel.mrStabilizingSteeringAngle;
	end	
	
	
	--intitiate mr variables	
	wheel.mrCurrentRollingResistance = 0.02;	
	wheel.mrLastTireLoad = wheel.restLoad;
	wheel.mrLastLongSlip = 1;
	wheel.mrRealLongSlip = 1;
	wheel.mrLastLatSlip = 0.35;	
	wheel.mrAvgLongSlip = 1;
	wheel.mrLastA = 0; --see vehicle "r, g, b, a = Utils.getTireTrackColorFromDensityBits(densityBits)" or "r, g, b, a, t = getTerrainAttributesAtWorldPos(g_currentMission.terrainRootNode, wx, wy, wz, true, true, true, true, false);"
	wheel.mrIsOnField = false;
	wheel.mrLastFrictionFx = 1;
	
end;
Vehicle.loadDynamicWheelDataFromXML = Utils.appendedFunction(Vehicle.loadDynamicWheelDataFromXML, Vehicle.mrLoadDynamicWheelDataFromXML);


--************************************************************************************************************************************************************
Vehicle.mrUpdateWheelTireFriction = function(self, wheel)	
	
	if self.isServer and self.isAddedToPhysics then
	
		if wheel.mrNotAWheel then				
			setWheelShapeTireFriction(wheel.node, wheel.wheelShape, 10*wheel.mrLongitudinalStiffnessScale, 100*wheel.mrLateralStiffnessScale, 141*wheel.mrLateralStiffnessScale, wheel.frictionScale*0.5);		
		else
		
			--for a farming game, lateral stiffness and longitudinal stiffness should not be "bound" like it is presently
			--especially because the "still" longitudinal slip = 20% and the "still" lateral slip = 20 (at still or in a straight line on a flat ground)
			--with a "normal" lateral stiffness, since the still lateral slip = 20, a "great" part of the longitudinal force is "capped" (limited) by the lateral force.
			--so, we have to "play" with "tricks" to get an acceptable result (not so acceptable ?)
			
			--another problem = longitudinal slip = 20% when the vehicle does not move, even if the wheels are slipping like hell. And so, we don't get the longitudinal force equivalent to 100% slip like it should
				
			local frictionFx = wheel.tireGroundFrictionCoeff*wheel.frictionScale*0.5; --0.5 because default wheel frictionScale = 2 in FS17
			local longStiff = 0.707 * wheel.mrLastTireLoad; -- 0.707 = 1/sqrt(2) == sqrt(2)/2
			local latStiff = wheel.mrLastTireLoad;		
			
			if self.lastSpeedReal>0.004 then --14.4kph
				local fx = (self.lastSpeedReal-0.004)*100;
				latStiff = latStiff * (1 + 25*fx)^0.5;
				--longStiff = longStiff * (1 + fx);
			end
			
			if self.lastSpeedReal>0.005 then -- 18kph
				local fx = 1 + math.min(2, (self.lastSpeedReal-0.005)*150);
				frictionFx = frictionFx * fx;			
			end
			
			if self.rotatedTime~=nil and self.rotatedTime~=0 then			
				if wheel.rotSpeed~=0 then
					--latStiff = latStiff * (1 + 2*math.abs(self.rotatedTime));
					latStiff = latStiff * math.max(1, 1 + 4*math.abs(self.rotatedTime) - self.lastSpeedReal*300); --2017-02-16
					--frictionFx = frictionFx * math.max(1, 1 + math.abs(self.rotatedTime) - self.lastSpeedReal*150);	--2017-02-16
					frictionFx = frictionFx * math.max(1, 1 + 0.5*math.abs(self.rotatedTime) - self.lastSpeedReal*200);	--2017-04-02
					--print("test rotated time ="..tostring(self.rotatedTime));
				end
			end
			
			local latStiffLoad = 1.414 * latStiff; -- sqrt(2) = 1.414	

			wheel.mrLastFrictionFx = frictionFx;
			
			setWheelShapeTireFriction(wheel.node, wheel.wheelShape, longStiff*wheel.mrLongitudinalStiffnessScale, latStiff*wheel.mrLateralStiffnessScale, latStiffLoad, frictionFx);
		end
				
	end
		
end;

--cheat the application force point for wheelshape so that we can use correct center of gravity for vehicles without having them to flip everytime they are cornering
Vehicle.mrUpdateWheelBase = function(self, wheel)

	if self.isServer and self.isAddedToPhysics and self.mrIsMrVehicle then	
		local xResult = wheel.positionX;
		if wheel.positionX>0 then
			xResult = xResult + wheel.mrForcePointOffsetX;
		else
			xResult = xResult - wheel.mrForcePointOffsetX;
		end;		
		--print("test mrUpdateWheelBase - wheel.mrForcePointOffsetX="..tostring(wheel.mrForcePointOffsetX) .. " - wheel.mrForcePointOffsetY="..tostring(wheel.mrForcePointOffsetY) .. " - resultX="..tostring(xResult) .. " - resultY="..tostring(wheel.positionY-wheel.mrForcePointOffsetY));
		setWheelShapeForcePoint(wheel.node, wheel.wheelShape, xResult, wheel.positionY-wheel.mrForcePointOffsetY, wheel.positionZ);		
	end
	
end
Vehicle.updateWheelBase = Utils.appendedFunction(Vehicle.updateWheelBase, Vehicle.mrUpdateWheelBase);

--************************************************************************************************************************************************************
--compute needed friction factor to achieve the desired longitudinal force
Vehicle.mrManageWheelsFriction = function(self, dt)
	
	--local frictionAdjust = 0;
	local totalMotorizedWheels = 0;	
	local totalDrivenWheelsSpeed = 0;	
	local numDrivenWheels = 0;
	
	--FOR DEBUG	--	
	local totalWheelSlip = 0;
	local sumGroundFrictionCoeff = 0;
	local totalLoadOnTires = 0;
	--END FOR DEBUG --	
			
	for _,wheel in ipairs(self.wheels) do
		local wheelSpd = wheel.mrLastWheelSpeed*wheel.radius;
		if wheel.mrIsDriven then
			totalDrivenWheelsSpeed = totalDrivenWheelsSpeed + wheelSpd;
			numDrivenWheels = numDrivenWheels + 1;
		end	
	
		if not wheel.mrNotAWheel and wheel.hasGroundContact then		
			--get current load on the wheel
			local tireLoad = getWheelShapeContactForce(wheel.node, wheel.wheelShape);
			if tireLoad ~= nil then					
				local nx,ny,nz = getWheelShapeContactNormal(wheel.node, wheel.wheelShape);
				local dx,dy,dz = localDirectionToWorld(wheel.node, 0,-1,0)
				wheel.mrLastTireLoad = math.max(0, -tireLoad*Utils.dotProduct(dx,dy,dz, nx,ny,nz)); -- 20170311 - protection to avoid negative value for the wheel load (amazone cayron for example)
				wheel.mrLastTireLoad = wheel.mrLastTireLoad + math.max(ny*9.81, 0.0) * wheel.mass; --add gravity force of tire	
				if Vehicle.debugRendering then
					wheel.mrLastContactForce = tireLoad;
					totalLoadOnTires = totalLoadOnTires	+ tireLoad;	
				end
			else				
				wheel.mrLastTireLoad = 0;
				wheel.mrLastContactForce = 0;
			end;
						
			if self.motorizedNode then	
				wheel.mrRealLongSlip = 1;
				--compute our longitudinal slip (formula = 1 - groundSpeed/wheelSpeed)
								
				local absWheelSpd = math.abs(wheelSpd);
				if self.movingDirection~=0 and self.movingDirection*wheelSpd<0 then
					absWheelSpd = 0;
				end				
				if absWheelSpd>0 then
					if absWheelSpd>=self.lastSpeedReal*1000 then
						wheel.mrRealLongSlip = 1 - self.lastSpeedReal*1000/absWheelSpd;
					elseif self.lastSpeedReal>0.0001 then --0.36kph
						wheel.mrRealLongSlip = 1 - absWheelSpd/(self.lastSpeedReal*1000);
					end					
				end

				--if (wheel.tireGroundFrictionCoeff*1.2)<1 then
				--	frictionAdjust = frictionAdjust + (1.2-0.2*wheel.mrRealLongSlip^1.5);
				--else
					--frictionAdjust = frictionAdjust + 1;
				--end;	

				wheel.mrAvgLongSlip	= wheel.mrAvgLongSlip * 0.95 + wheel.mrRealLongSlip * 0.05;
				
				
				totalMotorizedWheels = totalMotorizedWheels + 1;
				
				---------- debug ----------------
				if Vehicle.debugRendering then
					
					--update longitudinal slip
					local longSlip, latSlip = getWheelShapeSlip(wheel.node, wheel.wheelShape);
					if longSlip ~= nil then
						if self.lastSpeedReal<0.0001 then -- 0.36kph
							longSlip = 1;
						end;
						wheel.mrLastLongSlip = longSlip;	-- longSlip is always "positive"				
					else
						wheel.mrLastLongSlip = 1;				
					end;	

					--update lateral slip
					if latSlip ~= nil then
						wheel.mrLastLatSlip = latSlip;				
					else
						wheel.mrLastLatSlip = 0.35;				
					end;
						
					totalWheelSlip = totalWheelSlip + wheel.mrRealLongSlip;				
					sumGroundFrictionCoeff = sumGroundFrictionCoeff + wheel.tireGroundFrictionCoeff;
				end--end debug
				
			end
			
		else
		
			--no contact			
			wheel.mrLastContactForce = 0;
			wheel.mrLastTireLoad = 0;	
			wheel.mrAvgLongSlip	= 0;
			wheel.mrRealLongSlip = 1;			
		end--wheel contact			
		
	end --foreach wheel
	
	--if totalMotorizedWheels>0 then
		--self.mrWheelsFrictionFactor = frictionAdjust/totalMotorizedWheels;
	--end;
	
	--more friction when speed increase to help physx engine with heavy load at "high" speed
	--[[
	self.mrWheelsFrictionFactor = 1;
	if self.lastSpeedReal>0.005 then -- 18kph
		if self.attacherVehicle then
			self.mrWheelsFrictionFactor = 1 + math.min(2, 30*(self.lastSpeedReal-0.005));
		else
			self.mrWheelsFrictionFactor = 1 + math.min(2, 60*(self.lastSpeedReal-0.005));
		end
	end--]]
	if numDrivenWheels>0 then
		self.mrAvgDrivenWheelsSpeed = totalDrivenWheelsSpeed / numDrivenWheels;
	end
	
	for _,wheel in ipairs(self.wheels) do
		if not wheel.mrNotAWheel then					
			if wheel.contact ~= Vehicle.WHEEL_NO_CONTACT then
				local groundType = WheelsUtil.getGroundType(wheel.mrIsOnField, wheel.contact ~= Vehicle.WHEEL_GROUND_CONTACT, wheel.mrLastA);
				--local coeff = WheelsUtil.getTireFriction(wheel.tireType, groundType, g_currentMission.environment.groundWetness);
				local coeff = WheelsUtil.mrGetTireFriction(wheel, groundType, g_currentMission.environment.groundWetness);
				--if coeff ~= wheel.tireGroundFrictionCoeff then
				wheel.tireGroundFrictionCoeff = coeff;						
				--end
			end					
			self:updateWheelTireFriction(wheel);
			------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
		end
	end
	
	
	
	------------------------------------------------------------------------------------------------------------------------------
	--- DEBUG
	if Vehicle.debugRendering and self.isEntered and self.isServer and self.isActive and self:getIsActiveForInput() then
		
		--update total load on tire
		self.mrTotalWeightOnTires = totalLoadOnTires;
		
		--update avg wheel slip
		if totalMotorizedWheels>0 then
			self.mrAvgWheelSlip = self.mrAvgWheelSlip*0.98 + 0.02 * totalWheelSlip/totalMotorizedWheels;	
			self.mrAvgWheelGroundFrictionCoeff = sumGroundFrictionCoeff/totalMotorizedWheels;
		else
			self.mrAvgWheelSlip = 0;
			self.mrAvgWheelGroundFrictionCoeff = 0;
		end	

		--local axlePower = self.motor.mrLastTransmissionTorque*self.motor.clutchRpm/9.5488;
		local axlePower = self.motor.mrLastAxleTorque*self.motor.clutchRpm/9.5488;
		
		
		--expected slip function of TE, resistance forces, real speed and axle power
		--TE = tractive efficiency
		--Pdb = drawbar power
		--Pa = axle power
		--Fdb = drawbar force
		--Fr = resistance forces
		
		--TE = Pdb/Pa
		--Pdb = realSpd * Fdb
		--1. Fdb = Pdb / realSpd = TE*Pa / realSpd
		--2. Pa = wheelSpd * longForce => wheelSpd = Pa / longForce
		--3.longforce = Fdb + Fr
		--4. 2+3 =>  wheelSpd = Pa / (Fdb + Fr)
		--5. 4+1 => wheelSpd = Pa / [(TE * Pa / realSpd) + Fr]
		--6. Slip = 1 - realSpd / wheelSpd
		--7. 6+5 => Slip = 1 - (realSpd / Pa) * [(TE * Pa / realSpd) + Fr] = 1 - (TE + Fr * realSpd / Pa)
		
		--expected slip with Tractive Efficiency = 0.5
		local expectedSlipTE1 = 1;				
		--expected slip with Tractive Efficiency = 0.6
		local expectedSlipTE2 = 1;		
		--expected slip with Tractive Efficiency = 0.7
		local expectedSlipTE3 = 1;
		
		if axlePower>0 and self.mrDebugLastTotalResistance then
			--expected slip with Tractive Efficiency = 0.5
			expectedSlipTE1 = 1 - (0.5 + self.mrDebugLastTotalResistance*self.lastSpeedReal*1000/axlePower);					
			--expected slip with Tractive Efficiency = 0.6
			expectedSlipTE2 = 1 - (0.6 + self.mrDebugLastTotalResistance*self.lastSpeedReal*1000/axlePower);			
			--expected slip with Tractive Efficiency = 0.7
			expectedSlipTE3 = 1 - (0.7 + self.mrDebugLastTotalResistance*self.lastSpeedReal*1000/axlePower);
		end
		
		--slip = 1 - (TE + Fr*realSpd / Pa)
		--1-slip = TE + Fr*realSpd/Pa
		--TE = 1-slip - Fr*realSpd/Pa
		local TE = 0; 
		if self.mrDebugLastTotalResistance and axlePower>0 then
			TE = 1-self.mrAvgWheelSlip-self.mrDebugLastTotalResistance*self.lastSpeedReal*1000/axlePower;
		end
		
		--print(string.format("     mrWheelsFrictionFactor = %1.3f, sumLongForce=%1.3f, sumLongForceWanted=%1.3f, mrAvgWheelSlip=%1.3f",self.mrWheelsFrictionFactor,sumLongForce,sumLongForceWanted,self.mrAvgWheelSlip));
				
		--display expected slip (only ok on flat surface and if weight is perfectly dispatched on the tractor wheels according to the torqueRatio of differentials)
		local expectedSlip1 = 0;
		local expectedSlip2 = 0;
		
		if self.mrTotalWeightOnTires>0 and self.mrAvgWheelGroundFrictionCoeff>0 and self.mrDebugLastTotalResistance then
			local tractiveForce = self.mrDebugLastTotalResistance;		
			for _, implement in pairs(self.attachedImplements) do
				if implement.object and implement.object.mrDebugPowerConsumerLastForce then
					tractiveForce = tractiveForce + implement.object.mrDebugPowerConsumerLastForce;
				end
			end
			expectedSlip1 = (-1/6)*math.log(math.max(0.0024788,1-tractiveForce/(self.mrAvgWheelGroundFrictionCoeff*self.mrTotalWeightOnTires)));
			expectedSlip2 = (-1/5)*math.log(math.max(0.0024788,1-tractiveForce/(self.mrAvgWheelGroundFrictionCoeff*self.mrTotalWeightOnTires)));
		end		
		
		--display expected speed (only ok on flat surface and if weight is perfectly dispatched on the tractor wheels according to the torqueRatio of differentials)
		--P = F * Spd => axle power = tireForce * wheelSpd => wheelSpd = axlepower / tireForce
		--1 - (spd/wheel spd) = slip => spd = (1-slip)*wheel spd
		-- spd = (1-slip)*axlepower/tireforce
		
		local expectedSpd1 = 0;
		local expectedSpd2 = 0;
		if self.mrTotalWeightOnTires>0 and self.mrAvgWheelGroundFrictionCoeff>0 then
			if expectedSlip1>0 then				
				expectedSpd1 = (1-expectedSlip1)*axlePower/(self.mrTotalWeightOnTires*self.mrAvgWheelGroundFrictionCoeff*(1-math.exp(-6*expectedSlip1)));
			end
			if expectedSlip2>0 then				
				expectedSpd2 = (1-expectedSlip2)*axlePower/(self.mrTotalWeightOnTires*self.mrAvgWheelGroundFrictionCoeff*(1-math.exp(-5*expectedSlip2)));
			end
		end
		
		
		--we can use this paramter with the console commande "mrsetvalue2 value" to move the z position of the center of mass of the tractor
		if RealisticUtils.mrDebugValue2 then
			setCenterOfMass(self.components[1].node, self.components[1].centerOfMass[1], self.components[1].centerOfMass[2], self.components[1].centerOfMass[3]+tonumber(RealisticUtils.mrDebugValue2));
		end;
		
		if self.mrDebugLastExpectedSlip1==nil then self.mrDebugLastExpectedSlip1 = 0.5; end;
		self.mrDebugLastExpectedSlip1 = 0.99*self.mrDebugLastExpectedSlip1 + 0.01*expectedSlip1;
		
		if self.mrDebugLastExpectedSlip2==nil then self.mrDebugLastExpectedSlip2 = 0.5; end;
		self.mrDebugLastExpectedSlip2 = 0.99*self.mrDebugLastExpectedSlip2 + 0.01*expectedSlip2;
		
		
		renderText(0.02, 0.030, getCorrectTextSize(0.02), string.format("Expected speed min/max: %2.2f/%2.2f", expectedSpd2*3.6,expectedSpd1*3.6));
		renderText(0.02, 0.046, getCorrectTextSize(0.02), string.format("Expected slip min/max : %2.2f/%2.2f", self.mrDebugLastExpectedSlip1*100,self.mrDebugLastExpectedSlip2*100));
		
		renderText(0.02, 0.062, getCorrectTextSize(0.02), string.format("Slip TE 0.5/0.6/0.7= %2.2f/%2.2f/%2.2f", expectedSlipTE1*100,expectedSlipTE2*100,expectedSlipTE3*100));
		--renderText(0.02, 0.078, getCorrectTextSize(0.02), string.format("TE (Flat surface)= %2.2f", TE));
		
		renderText(0.2, 0.59, getCorrectTextSize(0.02), string.format("tractive Efficiency (flat surface): %2.2f", TE));	
		
		renderText(0.31, 0.46, getCorrectTextSize(0.02), "Avg Slip");
		renderText(0.36, 0.46, getCorrectTextSize(0.02), "Contact");
		renderText(0.41, 0.46, getCorrectTextSize(0.02), "Weight");
		renderText(0.45, 0.46, getCorrectTextSize(0.02), "Spd");
		local i = 0;
		for _,wheel in ipairs(self.wheels) do
			i = i +1;
			local wheelGroundSpeed = getWheelShapeAxleSpeed(wheel.node, wheel.wheelShape) * wheel.radius * 3.6;
			if self.mrVehicleIsStill then wheelGroundSpeed=0; end
			renderText(0.31, 0.46-i*0.022, getCorrectTextSize(0.02), string.format("%4.2f", wheel.mrAvgLongSlip*100));
			renderText(0.36, 0.46-i*0.022, getCorrectTextSize(0.02), string.format("%4.2f", wheel.mrLastContactForce));
			renderText(0.41, 0.46-i*0.022, getCorrectTextSize(0.02), string.format("%4.2f", wheel.mrLastTireLoad));
			renderText(0.45, 0.46-i*0.022, getCorrectTextSize(0.02), string.format("%4.2f", wheelGroundSpeed));
		end
		
	end;--end debug rendering
				
end


--MR ****************************************************************************************************
--return the "torque ratio" of the wheel
Vehicle.mrComputeWheelFinalTorqueRatio = function(self, wheel)
	
	if not self.differentials then
		return 1;
	end;
	
	local getFinalTorqueRatioFromDiffIndex;
	getFinalTorqueRatioFromDiffIndex = function(self, diffIndex)		
		for i,diff in pairs(self.differentials) do
			if not diff.diffIndex1IsWheel then
				if diff.diffIndex1==diffIndex then
					return diff.torqueRatio * getFinalTorqueRatioFromDiffIndex(self, i-1);
				end
			end
			if not diff.diffIndex2IsWheel then
				if diff.diffIndex2==diffIndex then
					return (1-diff.torqueRatio) * getFinalTorqueRatioFromDiffIndex(self, i-1);
				end
			end
		end
		return 1;
	end	
	
	for i,diff in pairs(self.differentials) do
		if diff.diffIndex1IsWheel then			
			--find the vehicle wheel
			if wheel==self.wheels[diff.diffIndex1] then							
				--compute the wheel torque ratio
				return diff.torqueRatio * getFinalTorqueRatioFromDiffIndex(self, i-1);				
			end	
		end
		if diff.diffIndex2IsWheel then			
			--find the vehicle wheel
			if wheel==self.wheels[diff.diffIndex2] then							
				--compute the wheel torque ratio
				return (1-diff.torqueRatio) * getFinalTorqueRatioFromDiffIndex(self, i-1);				
			end
		end		
	end
	
	return 1;

end



--MR ****************************************************************************************************
--update the wheels "mrIsDriven" parameter
Vehicle.mrUpdateWheelsIsDriven = function(self)
	
	for _,wheel in ipairs(self.wheels) do
		wheel.mrIsDriven = false;
		if self.differentials then
			for i,diff in pairs(self.differentials) do
				if diff.diffIndex1IsWheel then
					if wheel==self.wheels[diff.diffIndex1] then
						wheel.mrIsDriven = true;
						break;
					end
				end
				if diff.diffIndex2IsWheel then
					if wheel==self.wheels[diff.diffIndex2] then
						wheel.mrIsDriven = true;
						break;
					end
				end
			end
		end
	end

end

--MR ****************************************************************************************************
--Add wheel mass to the total mass
Vehicle.mrGetTotalMass = function(self, superFunc, addAttachables, addOnlyNonWheelAttachables)
	local mass = 0;
	if self.isServer then
		for _, comp in pairs(self.components) do
			mass = mass + getMass(comp.node);
		end;
		--MR add wheel mass
		for _, wheel in pairs(self.wheels) do
			mass = mass + wheel.mass;
		end
	else
		mass = self.serverMass;
		for _,v in pairs(self.specializations) do
			if v.getAdditiveClientMass ~= nil then
				mass = mass + v.getAdditiveClientMass(self);
			end;
		end;
	end;

	if addAttachables then
		if addOnlyNonWheelAttachables == nil then
			addOnlyNonWheelAttachables = true;
		end
		for _, implement in pairs(self.attachedImplements) do
			if implement.object ~= nil and (table.getn(implement.object.wheels) == 0 or not addOnlyNonWheelAttachables) then
				mass = mass + implement.object:getTotalMass(addAttachables, addOnlyNonWheelAttachables);
			end;
		end
	end

	return mass;
end;
Vehicle.getTotalMass = Utils.overwrittenFunction(Vehicle.getTotalMass, Vehicle.mrGetTotalMass);	


--******************************************************************************
--** return wheel width + additionnal wheel width (so, for a dual 710 setup, it returns 1.42
Vehicle.mrGetWheelTotalContactWidth = function(wheel)

	local width = wheel.width;
	if wheel.additionalWheels then
		for _,w in pairs(wheel.additionalWheels) do
			width = width + w.width;
		end
	end
	if wheel.tireType==WheelsUtil.getTireType("crawler") then
		wheel.mrIsCrawler = true;
	else
		wheel.mrIsCrawler = false;
	end
	return width;
	
end



Vehicle.testWheelFrictionForce = function(self, wheel)

	if true then return true end

	local tireLoad = getWheelShapeContactForce(wheel.node, wheel.wheelShape);
	if tireLoad==nil then
		tireLoad = wheel.mass * 9.81;
	else		
		tireLoad = tireLoad + wheel.mass * 9.81; --tireLoad in Kn
	end;

	local longMaxSlip = 1;
	local latMaxSlip = 0.9;

	local sizeX = 0.20;
	local sizeY = 0.20;
	
	local x = 0.7;
	local y = 0.5;

	local numGraphValues = 101;

	local longGraph = wheel.mrDebugLongitudinalFrictionGraph1;
	if longGraph == nil then
		longGraph = Graph:new(numGraphValues, x, y, sizeX, sizeY, 0, 0.0001, true, "", Graph.STYLE_LINES);
		longGraph:setColor(1, 1, 1, 1);
		wheel.mrDebugLongitudinalFrictionGraph1 = longGraph;
	end
	
	if Vehicle.mrDebugTestFxForSlip==nil then
		Vehicle.mrDebugTestFxForSlip=0;
	end;

	longGraph.maxValue = 1;
	local fx2=0.410 + Vehicle.mrDebugTestFxForSlip*0.005;
	
	local slip = 1;
	tireLoad = 10;
	local frictionFx = 1;
	
	fx2 = 0.1878650959*slip^2 -0.4373554613*slip + 0.4573023113;
	
	fx2 = 0.0454666;
	local latSlip = 0.35;--0.0000000001; 0.35 = lateral slip when going in a straight line on a plane road
	
	--longStiffness = 0.5*baseFrictionFx*frictionFx*tireLoad;
	
	setWheelShapeTireFriction(wheel.node, wheel.wheelShape, fx2*tireLoad, 10*tireLoad, 20*tireLoad, frictionFx);
	local longForce, latForce = computeWheelShapeTireForces(wheel.node, wheel.wheelShape, slip, latSlip, tireLoad);
	
	print("    slip=" .. tostring(slip) .. " fx = " .. tostring(fx2) .. " long force="..tostring(longForce/tireLoad) .. " - lat Force="..tostring(latForce/tireLoad));
	
	--[[
	for s=1, numGraphValues do
		local slip = (s-1)/(numGraphValues-1)* longMaxSlip;
		
		local groundFriction = 0.75;
		local frictionFx = groundFriction * math.max(1, 2 - math.sqrt(slip));
		local longStiff = 0.5*groundFriction*frictionFx*tireLoad;
		
		--setWheelShapeTireFriction(wheel.node, wheel.wheelShape, longStiff, 10*tireLoad, 20*tireLoad, frictionFx);
		
		--local fx = 0.5-0.08*slip;
		local fx = 0.7-0.25*math.exp(2*slip);-- + (1-slip^2);
		--fx = 0.9975;	

		
		fx2 = 0.1878650959*slip^2 -0.4373554613*slip + 0.4573023113;
		
		--setWheelShapeTireFriction doesn't seem to be taken into account if called multiple times here ???
		setWheelShapeTireFriction(wheel.node, wheel.wheelShape, fx2*tireLoad, 10*tireLoad, 20*tireLoad, 0.75);	
		
		local longForce, _ = computeWheelShapeTireForces(wheel.node, wheel.wheelShape, slip, 0.0000000001, tireLoad);
		--longGraph:setValue(s, longForce/tireLoad);	
		if self.testestest1==50 and (s<=6 or s==11 or s==16 or s==21 or s==31 or s==41 or s==51 or s==61 or s==71 or s==81 or s==91 or s==101) then 
			--print(string.format("  longGraph - slip=%3.4f frictionFx=%1.3f - longStiff=%1.3f - force=%2.4f", 100*slip, frictionFx, longStiff/tireLoad, longForce/tireLoad));
			print("    slip=" .. tostring(slip) .. " fx = " .. tostring(fx2) .. " force="..tostring(longForce/tireLoad));
			
		end;
		
		--longGraph.maxValue = math.max(longGraph.maxValue, longForce);
	end--]]
	
	if self.testestest1==50 then
		Vehicle.mrDebugTestFxForSlip = Vehicle.mrDebugTestFxForSlip + 1;
	end;
	
	--longGraph:draw();
	
	-------------------------------------------------------------------------------------
	--slip in rad
	local latGraph = wheel.mrDebugLateralFrictionGraph1;
	if latGraph == nil then
		latGraph = Graph:new(numGraphValues, 0.3, y, sizeX, sizeY, 0, 0.0001, true, "", Graph.STYLE_LINES);
		latGraph:setColor(1, 0, 0, 1);
		wheel.mrDebugLateralFrictionGraph1 = latGraph;
	end

	latGraph.maxValue = 1;
	local loadKn = 20;
	for s2=1, numGraphValues do
		local _, latForce = computeWheelShapeTireForces(wheel.node, wheel.wheelShape, 0, (s2-1)/(numGraphValues-1)* latMaxSlip, loadKn);
		latGraph:setValue(s2, math.abs(latForce)/loadKn);	
		if self.testestest1==600 then print(string.format("  latGraph - slip angle degree=%3.4f - force=%2.4f", (180/math.pi)*(s2-1)/(numGraphValues-1), latForce/loadKn)); end;
		--latGraph.maxValue = math.max(latGraph.maxValue, latForce);
	end
	
	--latGraph:draw();
	if self.testestest1==nil then
		self.testestest1 = 1;
	else
		self.testestest1 = self.testestest1+1;	
		--print("test : " .. tostring(self.testestest1));
	end;
	
end;

Vehicle.drawDebugMassOnTheWheels = function(self)

	if table.getn(self.wheels) > 0 then
		local gravity = 9.81;
		local totalWheelLoad = 0;
		for i, wheel in ipairs(self.wheels) do
			local tireLoad = getWheelShapeContactForce(wheel.node, wheel.wheelShape);
			if tireLoad ~= nil then
				local nx,ny,nz = getWheelShapeContactNormal(wheel.node, wheel.wheelShape);
				local dx,dy,dz = localDirectionToWorld(wheel.node, 0,-1,0);
				tireLoad = -tireLoad*Utils.dotProduct(dx,dy,dz, nx,ny,nz);
				tireLoad = tireLoad + math.max(ny*gravity, 0.0) * wheel.mass; --add gravity force of tire
			else
				tireLoad = 0;
			end
			totalWheelLoad = totalWheelLoad + tireLoad;
		end
		
		renderText(0.02, 0.015, getCorrectTextSize(0.02), string.format("Total mass on wheels : %2.2f", totalWheelLoad/gravity));			

	end;

end;