
--************************************************************************************************************************************************************
--called by the overwritten Vehicle.update function
--update the tire rolling resistance as well

WheelsUtil.mrGetTireFriction = function(wheel, groundType, wetScale)
	if wetScale == nil then
		wetScale = 0;
	end
	local coeff = WheelsUtil.tireTypes[wheel.tireType].frictionCoeffs[groundType];
	local coeffWet = WheelsUtil.tireTypes[wheel.tireType].frictionCoeffsWet[groundType];
	
	--rolling resistance coeff
	local rrDry = WheelsUtil.tireTypes[wheel.tireType].mrRollingResistanceCoeffs[groundType];
	local rrWet = WheelsUtil.tireTypes[wheel.tireType].mrRollingResistanceCoeffsWet[groundType];
	local baseRr = rrDry + (rrWet-rrDry)*wetScale;
	
	local frictionCoeff = coeff + (coeffWet-coeff)*wetScale;
	
	--TEST
	--frictionCoeff = 0.70;
	--baseRr = 0.055;
	--groundType = WheelsUtil.GROUND_FIELD;
	
	--very simplified compared to IRL
	--we want smaller rolling resistance bonus (greater penalty) for wider tires on high friction surfaces
	--we want greater rolling resistance bonus (smaller penalty) for wider tires on low friction surfaces
	
	--WheelsUtil.GROUND_FIELD
	--WheelsUtil.GROUND_ROAD
	--WheelsUtil.GROUND_SOFT_TERRAIN
	--WheelsUtil.GROUND_HARD_TERRAIN
	
	
	if wheel.mrLastTireLoad<=0 or wheel.mrIsCrawler then --protection in case mrLastTireLoad returns a negative value (math.pow 0.8 is not possible for a negative number)
		wheel.mrCurrentRollingResistance = baseRr;	
	else
		if groundType==WheelsUtil.GROUND_FIELD or groundType==WheelsUtil.GROUND_SOFT_TERRAIN then
			--wider tires = less rolling resistance compared to narrow tires (soft surface)
			--wheel.mrCurrentRollingResistance = baseRr * (1.1 - 0.2*wheel.mrTotalContactWidth -0.1*wheel.radius);
			--local fx = 0.02*wheel.mrLastTireLoad/(wheel.radius*wheel.mrTotalContactWidth);		
			--wheel.mrCurrentRollingResistance = baseRr * math.max(0.8, 1.3 -0.3*wheel.mrTotalContactWidth -0.1*wheel.radius);
			
			--wheel.mrCurrentRollingResistance = baseRr * math.max(0.8, 0.02 * wheel.mrLastTireLoad / (wheel.radius * wheel.mrTotalContactWidth))^0.5;
			local fx = (0.012 * wheel.mrLastTireLoad / (wheel.radius * wheel.mrTotalContactWidth))^0.8; -- 5T radius=0.9m, width=0.65m => fx = 1
			wheel.mrCurrentRollingResistance = baseRr * (0.8 + 0.2 * fx); 
			
			--wheel.mrSlippingResistanceEnabled = true;
		else
			--narrow tires = less rolling resistance compared to wider tires (hard surface)
			--wheel.mrCurrentRollingResistance = baseRr * (0.95 + 0.1*wheel.mrTotalContactWidth/wheel.radius);			
			local fx = 0.02*wheel.radius*wheel.mrLastTireLoad/wheel.mrTotalContactWidth;
			wheel.mrCurrentRollingResistance = baseRr * math.max(0.8, 1.2 - 0.15 * fx);
			
			--wheel.mrSlippingResistanceEnabled = false;
		end
	end
	
	--if wheel.radius>0 and wheel.mrLastTireLoad>0 and frictionCoeff>0 then
	--	wheel.mrCurrentRollingResistance = baseRr * (0.9 + 0.15*wheel.mrTotalContactWidth/wheel.radius) * math.max(0.8, 1.2 -6*wheel.radius*wheel.mrTotalContactWidth/(wheel.mrLastTireLoad*frictionCoeff^2));
	--end
	
	--if Vehicle.debugRendering then
	--	print(string.format("mrGetTireFriction - frictionCoeff=%1.2f - baseRr=%1.4f - resultRr=%1.4f - width=%1.2f, radius=%1.2f, lastTireLoad=%2.2f", frictionCoeff, baseRr, wheel.mrCurrentRollingResistance, wheel.mrTotalContactWidth, wheel.radius, wheel.mrLastTireLoad));
	--end
	
	return frictionCoeff;
end


--************************************************************************************************************************************************************
--
WheelsUtil.mrUpdateWheelsPhysics = function(self, superFunc, dt, currentSpeed, acceleration, doHandbrake, requiredDriveMode)

	--do not interfer with "genuine" vehicle not using the "mr" transmission
	if not self.mrUseMrTransmission then return superFunc(self, dt, currentSpeed, acceleration, doHandbrake, requiredDriveMode);end;

	local accelerationPedal = 0;
	local brakePedal = 0;	
	
	if doHandbrake then
		brakePedal = 1;			
	else

		if not g_currentMission.missionInfo.stopAndGoBraking then

			self.hasStopped = Utils.getNoNil(self.hasStopped, false);
			self.nextMovingDirection = Utils.getNoNil(self.nextMovingDirection, 0);
			
			--print("test - self.mrAvgDrivenWheelsSpeed=" .. tostring(self.mrAvgDrivenWheelsSpeed));			

			--if math.abs(acceleration) < 0.001 and math.abs(self.lastSpeedAcceleration) < 0.0001 and math.abs(self.lastSpeedReal) < 0.0001 and math.abs(self.lastMovedDistance) < 0.001 then
			if math.abs(acceleration) < 0.001 and math.abs(self.mrAvgDrivenWheelsSpeed)<0.02 then -- 0.02 m/s == 0.072 kph
				self.hasStopped = true;
				accelerationPedal = 0;
				brakePedal = 1;			
				self.nextMovingDirection = 0; --MR
			--elseif math.abs(self.lastSpeedReal) > 0.0001 then
			elseif math.abs(self.mrAvgDrivenWheelsSpeed) > 0.03 then -- 0.03 m/s == 0.108kph
				self.hasStopped = false;
			end

			if self.hasStopped and math.abs(acceleration) > 0.001 then
				self.nextMovingDirection = acceleration;
			end
			if self.nextMovingDirection * acceleration > 0.001 then
				accelerationPedal = acceleration;
				brakePedal = 0;
			elseif not self.hasStopped then
				accelerationPedal = 0;
				if self.nextMovingDirection == 0 then
					brakePedal = 1;
				else
					--brakePedal = math.abs(acceleration);

					if math.abs(acceleration) < 0.001 then
						--if currentSpeed < self.motor.lowBrakeForceSpeedLimit or doHandbrake then
						if (currentSpeed*1000) < self.motor.mrMinSpeed then
							--auto park brake
							brakePedal = math.min(1, self.mrLastBrakePedal + dt/1500); --1.5s to fully engage the park brake												
						end
					else
						brakePedal = math.abs(acceleration);
					end

				end
			end

		else --g_currentMission.missionInfo.stopAndGoBraking == true
		
			--print("--tracking update wheel physics --" .. tostring(g_currentMission.time));

			local brakeAcc = false;
			--if (self.movingDirection*currentSpeed*Utils.sign(acceleration*self.reverserDirection)) < -0.0003 then      -- 0.0003 * 3600 = 1.08 km/h
			local accDir = Utils.sign(acceleration*self.reverserDirection);
			if (self.mrAvgDrivenWheelsSpeed*accDir) < -0.2 then --0.2 = 0.72kph
				-- do we want to accelerate in the opposite direction of the vehicle speed?
				brakeAcc = true;
				
				self.mrSAGBtakeOff = false;
				
				--print("   1. self.mrAvgDrivenWheelsSpeed*accDir<-0.2 =" .. tostring(self.mrAvgDrivenWheelsSpeed*accDir));
				
			elseif self.movingDirection==0 then
				self.mrSAGBtryInversionFailed = false;
				self.mrSAGBtryInversion = false;
				--print("   2. movingDirection==0");
			elseif self.lastSpeedReal>0.003 and self.movingDirection*accDir==-1 then -- 20170308 - 0.003=10.8kph, we also want to continue braking when driven wheels are "blocked" while braking (no ABS) and the vehicle speed is not reasonnable (avoid stopping braking and making the driven wheels slip like hell to go in the right direction)
				brakeAcc = true;
				self.mrSAGBtakeOff = false;
				--print("   3. speed>5.4kph and bad moving direction");
			elseif (self.mrAvgDrivenWheelsSpeed*accDir) < 0 then
				--in this case, normally we do not brake since we want the inversion of direction to happen, but, in some cases (hilly road, "high" gear ratio and heavy load) the tractor keeps changing from braking/not braking because it takes too much time to inverse the direction (and then, once the brakes are released, the tractor goes in the wrong direction before being able to take off
				--if not self.mrSAGBtakeOff then
				--	brakeAcc = true;
				--end
				if self.mrSAGBtryInversionFailed then --0.18kph in case of inversion failure, wait for complete stop before trying again
					--print("   4.1 bad driven wheel direction and last inversion failed");
					brakeAcc = true;					
				else
					--print("   4.1 bad driven wheel direction but no failed inversion, trying to inverse direction");
					self.mrSAGBtryInversion = true;
				end		
			else
				--print("   5 - nothing special - reset parameters");
				self.mrSAGBtryInversionFailed = false;
				self.mrSAGBtryInversion = false;
			end;
			
			if self.mrSAGBtryInversion and brakeAcc then
				self.mrSAGBtryInversion = false;
				self.mrSAGBtryInversionFailed = true;
			end

			if math.abs(acceleration) < 0.001 then				
				if (currentSpeed*1000) < self.motor.mrMinSpeed then
					--auto park brake
					brakePedal = math.min(1, self.mrLastBrakePedal + dt/1500); --1.5s to fully engage the park brake				
				end;
			else
				if not brakeAcc then
					--print("   **** accelerating ****");
					accelerationPedal = acceleration;
					brakePedal = 0;
				else
					--print("   **** braking ****");
					accelerationPedal = 0;
					brakePedal = math.abs(acceleration);
				end;
			end;
			
			
			--print("test - accelerationPedal="..tostring(accelerationPedal) .. " - brakePedal="..tostring(brakePedal) .." - wheelSpeed="..tostring(self.mrAvgWheelSpeed) .. " - brakeAcc="..tostring(brakeAcc));

		end
		
	end -- doHandBrake
	
	
	--braking is not "ON/OFF" IRL => smooth a little the response time of the braking system (especially useful when playing with a keyboard)
	if brakePedal>self.mrLastBrakePedal then
		brakePedal = math.min(brakePedal, self.mrLastBrakePedal + dt/500); --500ms to be able to fully brake from a "null brake position"  --20170305
	end
	self.mrLastBrakePedal = brakePedal;

	if g_currentMission.missionInfo.stopAndGoBraking then
		self:setBrakeLightsVisibility(math.abs(brakePedal) > 0 and currentSpeed > 0.0006);
	else
		self:setBrakeLightsVisibility(math.abs(brakePedal) > 0 and not self.hasStopped);
	end

	self:setReverseLightsVisibility(self.movingDirection < 0 and (currentSpeed > 0.0006 or accelerationPedal < 0) and self.reverserDirection == 1);

	
	--MR : we want to know when the vehicle is braking
	self.mrIsBraking = brakePedal>0;
	
	
	--MR : 2017 02 17 - smooth a little the acceleration pedal	
	if brakePedal>0 then
		self.mrLastWantedAcc = 0;
	else
		local smoothTime = 500;		
		if accelerationPedal*self.mrLastWantedAcc<0 then
			self.mrLastWantedAcc = 0;
		elseif math.abs(accelerationPedal)<math.abs(self.mrLastWantedAcc) then
			smoothTime = 100; -- very few time when releasing the gaz pedal
		end
		
		if accelerationPedal>self.mrLastWantedAcc then			
			self.mrLastWantedAcc = math.min(accelerationPedal, self.mrLastWantedAcc+dt/smoothTime);
		else
			self.mrLastWantedAcc = math.max(accelerationPedal, self.mrLastWantedAcc-dt/smoothTime);
		end
	end
	accelerationPedal = self.mrLastWantedAcc;
	
	
		
	local ptoRpmWanted = PowerConsumer.getMaxPtoRpm(self);
	
	--print("test call self.motor:mrUpdateVehicleProps - accelerationPedal="..tostring(accelerationPedal) .." - self.reverserDirection="..tostring(self.reverserDirection) .." - speed="..tostring(self.lastSpeedReal*1000) .. " - self.movingDirection="..tostring(self.movingDirection) .." - ptoRpmWanted="..tostring(ptoRpmWanted) .. " - dt = " .. tostring(dt));
	local torque, maxRotSpeed, gearRatio, clutchTorque, rotInertia, dampingRate = self.motor:mrUpdateVehicleProps(accelerationPedal, self.reverserDirection, self.lastSpeedReal*1000, self.movingDirection, ptoRpmWanted, dt);	
	
	--print("test setVehicleProps - torque="..tostring(torque) .. " - maxRotSpeed="..tostring(maxRotSpeed) .. " - gearRatio="..tostring(gearRatio) .. " - clutchTorque="..tostring(clutchTorque) .. " - rotInertia="..tostring(rotInertia).. " - dampingRate="..tostring(dampingRate));
	
	setVehicleProps(self.motorizedNode, torque, maxRotSpeed, gearRatio, clutchTorque, rotInertia, dampingRate);
	--setVehicleProps(self.motorizedNode, torque, maxRotSpeed * math.pi / 30, gearRatio, clutchTorque, rotInertia, dampingRate);


	local doBrake = brakePedal > 0;
	for _, implement in pairs(self.attachedImplements) do
		if implement.object ~= nil then
			if doBrake then
				implement.object:onBrake(brakePedal);
			else
				implement.object:onReleaseBrake();
			end
		end
	end

	for _, wheel in pairs(self.wheels) do
		WheelsUtil.mrUpdateWheelPhysics(self, wheel, doHandbrake, brakePedal)
	end
	
	
	

end
WheelsUtil.updateWheelsPhysics = Utils.overwrittenFunction(WheelsUtil.updateWheelsPhysics, WheelsUtil.mrUpdateWheelsPhysics);


--************************************************************************************************************************************************************
--called by WheelsUtil.mrUpdateWheelsPhysics
--set the correct brake force to the wheel function of the wheel radius (otherwise, if the same brakeForce is applied to all wheels, the "small" wheels of the vehicle would always "block" themselves before the "tall" wheels
-- Example : tractor = front wheels "block" whereas rear wheels never block
-- Example : combine harvester = at still on a slope, rear wheels are blocked by the brakeForce whereas the front wheel turn (brakes are applied !)
-- Only useful for "moreRealistic" vehicles since base game vehicles "cheat" with a "more-than" double friction coeff (too much friction for a wheel to be blocked, even with a great brakeforce)
WheelsUtil.mrUpdateWheelPhysics = function(self, wheel, handbrake, brakePedal)

	if self.isAddedToPhysics then	
		if wheel.mrNotAWheel then
			local damping = wheel.rotationDamping;
			if not wheel.hasGroundContact and damping==0 then
				damping = 0.03 * wheel.mass;
			end
			setWheelShapeProps(wheel.node, wheel.wheelShape, 0, 0, 0, damping);			
		else
		
			local noBrakeLimitation = false;
		
			local brakeForce = brakePedal*self.motor:getBrakeForce();
			local rotationDamping = wheel.rotationDamping;
			if handbrake and wheel.hasHandbrake then
				brakeForce = self.motor:getBrakeForce()*10;
			end	
			
			brakeForce = brakeForce*wheel.brakeFactor*wheel.radius;
			
			--trick to get more brakeForce at still (otherwise, in very steep slopes, the wheels can turn slowly downhill at still)
			if brakePedal==1 then
				if brakeForce==0 then
					if wheel.mrIsDriven then
						--rotationDamping = 0.1;
						rotationDamping = 0.5 * wheel.mass; --20170305 - takes into account wheel mass (same time to reach 0 rotSpeed whatever the mass of the wheel)
					end
				elseif self.lastSpeedReal<0.0001 then --0.36kph
					brakeForce = 3*brakeForce;
					noBrakeLimitation = true;
				end				
			end
			
			if brakeForce>0 and not noBrakeLimitation then			
				--20170308 - limit brake force according to weight on the wheel
				--brakeForce = Utils.clamp(1.1*wheel.mrLastTireLoad*wheel.radius/math.max(1, wheel.mrLastFrictionFx)^0.5, 0.1*brakeForce, brakeForce);
				brakeForce = Utils.clamp(1.1*wheel.mrLastTireLoad*wheel.radius/Utils.clamp(wheel.mrLastFrictionFx, 1, 1.6), 0.1*brakeForce, brakeForce);
			end
			
			--20170316 - add damping when no groundcontact
			if rotationDamping==0 and not wheel.hasGroundContact then
				rotationDamping = 0.03 * wheel.mass;
			end
			
			--[[
			--simulate rolling resistance at still
			if brakeForce==0 then
				local wheelGroundSpd = math.abs(wheel.mrLastWheelSpeed*wheel.radius);
				if wheelGroundSpd<0.2 then --0.2 = 0.72kph
					brakeForce = wheel.mrLastTireLoad * wheel.mrCurrentRollingResistance*wheel.radius*math.min(1, 2-10*wheelGroundSpd);
				end
			--else
				--"remove" the mrWheelsFrictionFactor benefit from the braking force
				--brakeForce = brakeForce / self.mrWheelsFrictionFactor;
			end--]]
			
			--if Vehicle.debugRendering and self.isControlled and self.motor then			
			--	local longSlip, latSlip = getWheelShapeSlip(wheel.node, wheel.wheelShape); 
			--	print(string.format("%1.2f WheelsUtil.mrUpdateWheelPhysics - brakeForce=%1.2f - radius=%1.2f - rotationDamping=%1.2f - wheelAxleSpd=%1.4f wheelGroundSpeed=%1.2f vehicleSpeed=%1.2f longSlip=%2.2f latSlip=%2.2f frictionFactor=%1.2f gearRatio=%2.4f clutchTorque=%3.3f clutchRpm=%4.2f nonClampedRpm=%4.2f",
			--	g_currentMission.time,brakeForce, wheel.radius, wheel.rotationDamping, wheel.mrLastWheelSpeed, wheel.mrLastWheelSpeed*wheel.radius*3.6, self.lastSpeedReal*3600, longSlip*100, latSlip*100, self.mrWheelsFrictionFactor, self.motor.gearRatio, 1000*self.motor.mrLastClutchTorque, self.motor.clutchRpm, self.motor.nonClampedMotorRpm));
			--end;
			local steeringAngle = wheel.steeringAngle;
			if steeringAngle==0 then
				steeringAngle = wheel.mrStabilizingSteeringAngle;
			end
			
			setWheelShapeProps(wheel.node, wheel.wheelShape, 0, brakeForce, steeringAngle, rotationDamping);	
		end
	end
end



--************************************************************************************************************************************************************
-- update mrVehicleIsStill
WheelsUtil.mrUpdateWheelsGraphics = function(self, superFunc, dt)

	--do not interfer with "genuine" vehicle
	if not self.mrIsMrVehicle then return superFunc(self, dt);end;
	
	--print(g_currentMission.time .. " mrUpdateWheelsGraphics for " .. self.configFileName .. " - self.isAddedToPhysics="..tostring(self.isAddedToPhysics));

	if self.isServer then
		self.hasWheelGroundContact = false;
	end
	
	--MR *****************	
	self.mrTotalxDriveDiff = 0;
	self.mrVehicleIsStill = false;
	----------------------

	for _, wheel in pairs(self.wheels) do
		WheelsUtil.updateWheelSteeringAngle(self, wheel, dt);
		if self.isServer and self.isAddedToPhysics then
			WheelsUtil.updateWheelHasGroundContact(wheel);
			if wheel.hasGroundContact then
				self.hasWheelGroundContact = true;
			end

			if wheel.updateWheel then
				local x,y,z, xDrive, suspensionLength = getWheelShapePosition(wheel.node, wheel.wheelShape);
				
				--print(tostring(g_currentMission.time) .. " test wheel update graphics - xDrive="..tostring(xDrive) .. " - xDriveOffset="..tostring(wheel.xDriveOffset) .. " - suspensionLength = " .. tostring(suspensionLength) .. " - wheel.node=" .. tostring(wheel.node) .. " - wheel.wheelShape="..tostring(wheel.wheelShape));
				
				xDrive = xDrive + wheel.xDriveOffset
				WheelsUtil.updateWheelGraphics(self, wheel, x, y, z, xDrive, suspensionLength);

				--fill netinfo (on server)
				wheel.netInfo.x = x;
				wheel.netInfo.y = y;
				wheel.netInfo.z = z;
				
				--MR ************************************				
				self.mrTotalxDriveDiff = self.mrTotalxDriveDiff + xDrive - wheel.netInfo.xDrive;				
				--******************************************
				
				wheel.netInfo.xDrive = xDrive;
				wheel.netInfo.suspensionLength = suspensionLength;
			else
				wheel.updateWheel = true
			end
		else
			-- client code
			local x, y, z = wheel.netInfo.x, wheel.netInfo.y, wheel.netInfo.z;
			local xDrive = wheel.netInfo.xDrive;
			local suspensionLength = wheel.netInfo.suspensionLength
			WheelsUtil.updateWheelGraphics(self, wheel, x, y, z, xDrive, suspensionLength);
		end
	end	
	
	--MR ******************
	if self.isServer then
		if not self.isAddedToPhysics or (self.lastSpeedReal==0 and self.mrTotalxDriveDiff==0) then	
			-- vehicle is at still and so, getWheelShapeContactForce/getWheelShapeContactNormal/getWheelShapeAxleSpeed/getWheelShapeSlip/getMotorRotationSpeed can return "fake" values (not reliable)
			self.mrVehicleIsStill = true;
			--print("test - mr vehicle is still : " .. tostring(self.configFileName));
		end
	end	
	--print("test - mrVehicleIsStill="..tostring(self.mrVehicleIsStill) .. " - self.lastSpeedReal="..tostring(self.lastSpeedReal*3600) .." - self.mrTotalxDriveDiff="..tostring(self.mrTotalxDriveDiff));
	-----------------------
	
end
WheelsUtil.updateWheelsGraphics = Utils.overwrittenFunction(WheelsUtil.updateWheelsGraphics, WheelsUtil.mrUpdateWheelsGraphics);


--************************************************************************************************************************************************************
-- modify tyre deformation script
WheelsUtil.mrUpdateWheelGraphics = function(self, superFunc, wheel, x, y, z, xDrive, suspensionLength)

	--do not interfer with "genuine" vehicle
	if not self.mrIsMrVehicle then return superFunc(self, wheel, x, y, z, xDrive, suspensionLength);end;

	local steeringAngle = wheel.steeringAngle;
	if not wheel.showSteeringAngle then
		steeringAngle = 0;
	end

	local dirX, dirY, dirZ = 0, -1, 0
	if wheel.repr == wheel.driveNode then
		setRotation(wheel.repr, xDrive, steeringAngle, 0);
	else
		dirX, dirY, dirZ = localDirectionToLocal(wheel.repr, getParent(wheel.repr), 0, -1, 0);
		setRotation(wheel.repr, 0, steeringAngle, 0);		
		setRotation(wheel.driveNode, xDrive, 0, 0);
	end

	if wheel.wheelTire ~= nil then
		local x, y, z, _ = getShaderParameter(wheel.wheelTire, "morphPosition");
		
		
		--not adequate because on some vehicle, we have to "cheat" with the deltaY to get the wheel at the right position (example : negative "initialCompression" value)
		--local deformation = Utils.clamp((wheel.deltaY+0.04-suspensionLength)*0.7, 0, wheel.maxDeformation)
		local deformation = 0;
		if wheel.mrTotalContactWidth>0 then
			--[[local deformationFx = Utils.clamp(-0.6+0.017*wheel.mrLastTireLoad*wheel.mrPressureFx/(wheel.mrTotalContactWidth*wheel.radius^0.5), 0, 1);	
			--deformationFx = Utils.clamp(-0.6+0.017*wheel.mrLastTireLoad*wheel.mrPressureFx/(wheel.radius*wheel.mrTotalContactWidth), 0, 1);		
			if deformationFx<=0.3 then
				deformation = 0;
			elseif deformationFx<=0.5 then
				deformation = 3.5*(deformationFx-0.3)*wheel.maxDeformation; -- 0 to 70%
			else
				--deformationFx between 0.5 and 1
				deformation = (0.7 + (deformationFx-0.5)*0.6)*wheel.maxDeformation;
			end--]]
			
			deformation =  Utils.clamp(-0.04+wheel.maxDeformation*0.1/wheel.mrTotalContactWidth/wheel.mrPressureFx*(wheel.mrLastTireLoad/wheel.radius)^0.5, 0, wheel.maxDeformation);	
			
			--if self.lastSpeedReal*3600>5 then
			--	print("test wheel - maxDeformation="..tostring(wheel.maxDeformation) .." - deformation="..tostring(deformation) .." - radius="..tostring(wheel.radius));
			--end
			
		end
		
		setShaderParameter(wheel.wheelTire, "morphPosition", x, y, z, deformation, false);

		if wheel.additionalWheels ~= nil then
			for _, additionalWheel in pairs(wheel.additionalWheels) do
				local x, y, z, _ = getShaderParameter(additionalWheel.wheelTire, "morphPosition");
				setShaderParameter(additionalWheel.wheelTire, "morphPosition", x, y, z, deformation, false);
			end
		end

		suspensionLength = suspensionLength+deformation
	end

	suspensionLength = suspensionLength - wheel.deltaY
	setTranslation(wheel.repr, wheel.startPositionX + dirX*suspensionLength, wheel.startPositionY + dirY*suspensionLength, wheel.startPositionZ + dirZ*suspensionLength);

	if wheel.steeringNode ~= nil then
		local refAngle = wheel.steeringNodeMaxRot;
		local refTrans = wheel.steeringNodeMaxTransX
		local refRot = wheel.steeringNodeMaxRotY;
		if steeringAngle < 0 then
			refAngle = wheel.steeringNodeMinRot;
			refTrans = wheel.steeringNodeMinTransX;
			refRot = wheel.steeringNodeMinRotY;
		end
		local steering = 0;
		if refAngle ~= 0 then
			steering = steeringAngle / refAngle;
		end

		if wheel.steeringNodeMinTransX ~= nil then
			local x,y,z = getTranslation(wheel.steeringNode);
			x = refTrans * steering;
			setTranslation(wheel.steeringNode, x, y, z);
		end
		if wheel.steeringNodeMinRotY ~= nil then
			local rotX,rotY,rotZ = getRotation(wheel.steeringNode);
			rotY = refRot * steering;
			setRotation(wheel.steeringNode, rotX, rotY, rotZ);
		end
	end

	if wheel.fenderNode ~= nil then
		local angleDif = 0;
		if steeringAngle > wheel.fenderRotMax then
			angleDif = wheel.fenderRotMax - steeringAngle;
		elseif steeringAngle < wheel.fenderRotMin then
			angleDif = wheel.fenderRotMin - steeringAngle;
		end
		setRotation(wheel.fenderNode, 0, angleDif, 0);
	end
	
end
WheelsUtil.updateWheelGraphics = Utils.overwrittenFunction(WheelsUtil.updateWheelGraphics, WheelsUtil.mrUpdateWheelGraphics);
