RealisticUtils = {};

RealisticUtils.defaultVehiclesModifiedData = {};

RealisticUtils.nameToGroundType = {
	["GROUND_ROAD"] = WheelsUtil.GROUND_ROAD, 
	["GROUND_HARD_TERRAIN"] = WheelsUtil.GROUND_HARD_TERRAIN,
	["GROUND_SOFT_TERRAIN"] = WheelsUtil.GROUND_SOFT_TERRAIN,
	["GROUND_FIELD"] = WheelsUtil.GROUND_FIELD};

--**********************************************************************************************************************************************************
RealisticUtils.loadDefaultVehiclesModifiedData = function(folderPath, fileName)

	--reset vehicles data
	RealisticUtils.defaultVehiclesModifiedData = {};

	--loading database file
	local xmlFile = loadXMLFile("realisticDefaultVehicleDatabase.xml", folderPath .. "/" .. fileName);
	
	local i = 0;
	while true do
		local vehicleXmlPath = string.format("vehicles.vehicle(%d)", i);
		if not hasXMLProperty(xmlFile, vehicleXmlPath) then break; end;		
		local vehicleFilePath = getXMLString(xmlFile, vehicleXmlPath);	
		local overridedFileName = getXMLString(xmlFile, vehicleXmlPath .. "#overridedFileName");
		RealisticUtils.defaultVehiclesModifiedData[overridedFileName] = folderPath .. "/" .. vehicleFilePath;

		--print("RealisticUtils.loadDefaultVehiclesModifiedData - overridedFileName="..tostring(overridedFileName) .. " - vehicleFilePath="..tostring(folderPath .. "/" .. vehicleFilePath));
		
		--load the modified store data
		RealisticUtils.loadModifiedStoreData(overridedFileName, folderPath .. "/" .. vehicleFilePath);		
		
		i = i + 1;
	end	
	
	delete(xmlFile);

end


--**********************************************************************************************************************************************************
RealisticUtils.loadModifiedStoreData = function(overridedXmlFilePath, modifiedDataXmlFilePath)

	local genuineStoreItem = StoreItemsUtil.storeItemsByXMLFilename[overridedXmlFilePath:lower()];	

	if genuineStoreItem==nil then
		RealisticUtils.printWarning("RealisticUtils.loadModifiedStoreData", string.format("genuineStoreItem is nil for xml path=%s",overridedXmlFilePath), false);
		return;
	end
	
	local tmpModifiedStoreItem = StoreItemsUtil.loadStoreItem(modifiedDataXmlFilePath,nil,nil,false,false);
	
	--copy the modified data to the genuine store item
	genuineStoreItem.name = tmpModifiedStoreItem.name;
	genuineStoreItem.dailyUpkeep = tmpModifiedStoreItem.dailyUpkeep;
	genuineStoreItem.lifetime = tmpModifiedStoreItem.lifetime;
	genuineStoreItem.price = tmpModifiedStoreItem.price;
	genuineStoreItem.specs = tmpModifiedStoreItem.specs;	
	genuineStoreItem.configurations = tmpModifiedStoreItem.configurations;	
	genuineStoreItem.brand = tmpModifiedStoreItem.brand;	
	genuineStoreItem.imageBrand = tmpModifiedStoreItem.imageBrand;
	
	--delete the temp modified storeitem
	StoreItemsUtil.removeStoreItem(tmpModifiedStoreItem.id);
	
end


--**********************************************************************************************************************************************************
RealisticUtils.loadRealFillTypesData = function(filePath)
	
	local xmlFile = loadXMLFile("realFillTypesXML", filePath);
	
	local i = 0;
	while true do
		local fillTypeName = string.format("fillTypes.fillType(%d)", i);
		if not hasXMLProperty(xmlFile, fillTypeName) then break; end;
		
		local realFillType = {};
		realFillType.name = getXMLString(xmlFile, fillTypeName .. "#name");
		if realFillType.name == nil then
			RealisticUtils.printWarning("RealisticUtils.loadRealFillTypesData", "realFillType.name is nil, i="..tostring(i), true);
			break;
		end;
		realFillType.density = getXMLFloat(xmlFile, fillTypeName .. "#density");
		if realFillType.density == nil then
			RealisticUtils.printWarning("RealisticUtils.loadRealFillTypesData", "realFillType.density is nil, i="..tostring(i), true);
			break;
		end;
		realFillType.startPricePerM3 = getXMLFloat(xmlFile, fillTypeName .. "#startPricePerM3");
		
		local fillType = FillUtil.fillTypeNameToDesc[realFillType.name];
		if fillType==nil then
			RealisticUtils.printWarning("RealisticUtils.loadRealFillTypesData", "fillType is unknown on this map, name="..realFillType.name, false);
		else
			fillType.massPerLiter = 0.001 * realFillType.density;-- * 0.2;
			if realFillType.startPricePerM3~= nil then
				--print("fillType "..tostring(fillType.name).." - old startPrice="..tostring(fillType.startPricePerLiter*1000).." - old price="..tostring(fillType.pricePerLiter*1000) .. " - new start price="..tostring(realFillType.startPricePerM3));
				fillType.startPricePerLiter = 0.001 * realFillType.startPricePerM3;
				fillType.pricePerLiter = fillType.startPricePerLiter;				
			end;			
		end;
		
		i = i + 1;
	end
	
	delete(xmlFile);

end;


--**********************************************************************************************************************************************************
RealisticUtils.loadRealFruitTypesData = function(filePath)
	
	local xmlFile = loadXMLFile("realFruitTypesXML", filePath);
	
	local i = 0;
	while true do
		local fruitTypeName = string.format("fruitTypes.fruitType(%d)", i);
		if not hasXMLProperty(xmlFile, fruitTypeName) then break; end;
		
		local realFruitType = {};
		realFruitType.name = getXMLString(xmlFile, fruitTypeName .. "#name");
		if realFruitType.name == nil then
			RealisticUtils.printWarning("RealisticUtils.loadRealFruitTypesData", "realFruitType.name is nil, i="..tostring(i), true);
			break;
		end;
		realFruitType.literPerSqm = getXMLFloat(xmlFile, fruitTypeName .. "#literPerSqm");
		if realFruitType.literPerSqm == nil then
			RealisticUtils.printWarning("RealisticUtils.loadRealFruitTypesData", "realFruitType.literPerSqm is nil, i="..tostring(i), true);
			break;
		end;
		realFruitType.seedUsagePerSqm = getXMLFloat(xmlFile, fruitTypeName .. "#seedUsagePerSqm");
		if realFruitType.seedUsagePerSqm == nil then
			RealisticUtils.printWarning("RealisticUtils.loadRealFruitTypesData", "realFruitType.seedUsagePerSqm is nil, i="..tostring(i), true);
			break;
		end;
		realFruitType.windrowLiterPerSqm = getXMLFloat(xmlFile, fruitTypeName .. "#windrowLiterPerSqm");
		
		realFruitType.mrMaterialQtyFx = Utils.getNoNil(getXMLFloat(xmlFile, fruitTypeName .. "#mrMaterialQtyFx"), 1);
				
		local fruitType = FruitUtil.fruitTypes[realFruitType.name];
		if fruitType==nil then
			RealisticUtils.printWarning("RealisticUtils.loadRealFruitTypesData", "fruitType is unknown, name="..realFruitType.name, false);
		else
			fruitType.literPerSqm = realFruitType.literPerSqm;
			fruitType.seedUsagePerSqm = realFruitType.seedUsagePerSqm;
			if realFruitType.windrowLiterPerSqm~= nil then
				fruitType.windrowLiterPerSqm = realFruitType.windrowLiterPerSqm;
			end;
			fruitType.mrMaterialQtyFx = realFruitType.mrMaterialQtyFx;
		end;
		
		i = i + 1;
	end
	
	delete(xmlFile);

end;


--**********************************************************************************************************************************************************
RealisticUtils.loadRealTyresFrictionAndRr = function(filePath)
	
	--DebugUtil.printTableRecursively(WheelsUtil.GROUND_HARD_TERRAIN, 1, 1, 100);
	
	
	local xmlFile = loadXMLFile("realFrictionAndRrXML", filePath);
	
	local i = 0;
	while true do
		local tyreTypeKey = string.format("tyreTypes.tyreType(%d)", i);
		if not hasXMLProperty(xmlFile, tyreTypeKey) then break; end;
		
		local tyreTypeName = getXMLString(xmlFile, tyreTypeKey .. "#name");
		
		local tireTypeIndex = WheelsUtil.getTireType(tyreTypeName);
		
		if tireTypeIndex==nil then
			RealisticUtils.printWarning("RealisticUtils.loadRealTyresFrictionAndRr", "unknown tireType, tyreTypeName="..tostring(tyreTypeName) .. ", i="..tostring(i), true);
			break;
		end;
		
		local tireType = WheelsUtil.tireTypes[tireTypeIndex];
		
		--new table to store rolling resistance values
		tireType.mrRollingResistanceCoeffs = {};
		tireType.mrRollingResistanceCoeffsWet = {};
		
		--for each surface type, set the new values
		local j = 0;
		while true do
			local surfaceTypeKey = tyreTypeKey .. string.format(".surfaceType(%d)", j);
			if not hasXMLProperty(xmlFile, surfaceTypeKey) then break; end;
			
			local surfaceTypeName = getXMLString(xmlFile, surfaceTypeKey .. "#name");
			local groundType = RealisticUtils.nameToGroundType[surfaceTypeName];
			
			if groundType==nil then
				RealisticUtils.printWarning("RealisticUtils.loadRealTyresFrictionAndRr", "unknown groundType, surfaceTypeName="..tostring(surfaceTypeName) .. ", i="..tostring(i) .. ", j="..tostring(j), true);
				break;
			end;
			
			local getValueFromXML = function(xmlPath, myTable, groundType)				
				local value = getXMLFloat(xmlFile, xmlPath);
				if value==nil then
					RealisticUtils.printWarning("RealisticUtils.loadRealTyresFrictionAndRr", "nil value for groundType, xmlPath="..tostring(xmlPath), true);					
					return false;
				end;
				myTable[groundType] = value;
				return true;
			end;
			
			if not getValueFromXML(surfaceTypeKey .. "#dryFriction", tireType.frictionCoeffs, groundType) then break; end;
			if not getValueFromXML(surfaceTypeKey .. "#wetFriction", tireType.frictionCoeffsWet, groundType) then break; end;
			if not getValueFromXML(surfaceTypeKey .. "#dryRollingResistance", tireType.mrRollingResistanceCoeffs, groundType) then break; end;
			if not getValueFromXML(surfaceTypeKey .. "#wetRollingResistance", tireType.mrRollingResistanceCoeffsWet, groundType) then break; end;
			
			j = j + 1;
		end
		
		i = i + 1;
	end
	
	delete(xmlFile);

end;

--**********************************************************************************************************************************************************
RealisticUtils.getDragArea = function(self)
	
	local maskingTractorArea = math.max(self.mrDragArea, self.mrOverridedTrailerArea) ;
	local dragArea = self.mrDragArea;
	
	local trailerMaxDragArea = 0;
	for _, implement in pairs(self.attachedImplements) do
		if implement.object ~= nil and implement.object.mrIsMrVehicle then
			trailerMaxDragArea = math.max(trailerMaxDragArea, implement.object.mrDragArea);
		end;
	end
	
	if trailerMaxDragArea>maskingTractorArea then
		dragArea = dragArea + trailerMaxDragArea-maskingTractorArea;
	end	
	
	return dragArea;
end;

--fx must be between 0 and 1
RealisticUtils.linearFx = function(fx, minVal, maxVal)
	return minVal + fx * (maxVal-minVal);
end;

--return 1 if param=minParam, return minVal if param=maxParam
--param between minParam and maxParam
RealisticUtils.linearFx2 = function(param, minParam, maxParam, minVal)
	return 1 - (1-minVal)*(param - minParam)/(maxParam-minParam);
end;



--same as Utils.getFileName but check if we need to replace the baseDirectory with the mr one (tag = "$mr" or "$MR" to start the filename with)
RealisticUtils.getFilename = function(filename, baseDirectory)

	if string.sub(filename,1,3):lower()=="$mr" then			
		--print("test : " .. string.sub(filename,1,3):lower() .. " - 2= " .. string.sub(filename,4) .. " - 3=" ..RealisticUtils.mrEngineDirectory .. 
		--" - 4="..Utils.getFilename(string.sub(filename,4), RealisticUtils.mrEngineDirectory));
	
		return Utils.getFilename(string.sub(filename,5), RealisticUtils.mrEngineDirectory);		
	else
		return Utils.getFilename(filename, baseDirectory);		
	end	
	
end

--**********************************************************************************************************************************************************
-- allow us some modification of the i3d node at runtime
-- added 20170326 to "unlink" the collision shape of the markers of the Horsch Maestro
RealisticUtils.manageNodes = function(self, xmlFile)

	local nodeI = 0;
	while true do
		local nodeKey = string.format("vehicle.mrManageNodes.manageNode(%d)", nodeI);
		if not hasXMLProperty(xmlFile, nodeKey) then
			break;
		end;
		
		local node = Utils.indexToObject(self.components, getXMLString(xmlFile, nodeKey.."#node"));
		local unlinkNode = Utils.getNoNil(getXMLBool(xmlFile, nodeKey.."#unlink"), false);
		
		if unlinkNode then
			unlink(node);
		end
	
		nodeI = nodeI + 1;
	end --while
	
end


--**********************************************************************************************************************************************************
--***** TEST / DEBUG
--**********************************************************************************************************************************************************






RealisticUtils.printWarning = function(stackTrace, message, isError)
	
	local gameTime = 0;
	if g_currentMission then gameTime = g_currentMission.time; end
	local msg = "*** " .. tostring(gameTime) .. " MoreRealistic - ";
	--local msg = "*** MoreRealistic - ";
	
	if isError then
		msg = msg .. "ERROR - ";
	else
		msg = msg .. "WARNING - ";
	end;
	
	msg = msg .. stackTrace .. " - " .. message;
	
	print(msg);	

end;




function RealisticUtils.testClass(className, classToTest)

	print("testing " .. tostring(className));		
	for _,k in pairs(classToTest) do
		print("function name : " .. tostring(_) .. " value : " .. tostring(k));			
	end;
	
	--table.foreach(classToTest,print);
	
	print("end testing " .. tostring(className));
	
end;

function RealisticUtils.testObject(displayName, object)

	print("testing " .. tostring(displayName));
	for key, value in pairs(getmetatable(object)) do
		print(key, value);
	end;
	print("end testing " .. tostring(displayName));
end;


--comes from courseplay. Thanks Jakob !
function RealisticUtils.testTable(t, name, indent, maxDepth)
	
	local cart -- a container
	local autoref -- for self references
	maxDepth = maxDepth or 50;
	local depth = 0;

	--[[ counts the number of elements in a table
local function tablecount(t)
   local n = 0
   for _, _ in pairs(t) do n = n+1 end
   return n
end
]]
	-- (RiciLake) returns true if the table is empty
	local function isemptytable(t) return next(t) == nil end

	local function basicSerialize(o)
		local so = tostring(o)
		if type(o) == "function" then
			local info = debug.getinfo(o, "S")
			-- info.name is nil because o is not a calling level
			if info.what == "C" then
				return string.format("%q", so .. ", C function")
			else
				-- the information is defined through lines
				return string.format("%q", so .. ", defined in (" ..
						info.linedefined .. "-" .. info.lastlinedefined ..
						")" .. info.source)
			end
		elseif type(o) == "number" then
			return so
		else
			return string.format("%q", so)
		end
	end

	local function addtocart(value, name, indent, saved, field, curDepth)
		indent = indent or ""
		saved = saved or {}
		field = field or name
		cart = cart .. indent .. field

		if type(value) ~= "table" then
			cart = cart .. " = " .. basicSerialize(value) .. ";\n"
		else
			if saved[value] then
				cart = cart .. " = {}; -- " .. saved[value]
						.. " (self reference)\n"
				autoref = autoref .. name .. " = " .. saved[value] .. ";\n"
			else
				saved[value] = name
				--if tablecount(value) == 0 then
				if isemptytable(value) then
					cart = cart .. " = {};\n"
				else
					if curDepth <= maxDepth then
						cart = cart .. " = {\n"
						for k, v in pairs(value) do
							k = basicSerialize(k)
							local fname = string.format("%s[%s]", name, k)
							field = string.format("[%s]", k)
							-- three spaces between levels
							addtocart(v, fname, indent .. "\t", saved, field, curDepth + 1);
						end
						cart = cart .. indent .. "};\n"
					else
						cart = cart .. " = { ... };\n";
					end;
				end
			end
		end;
	end

	name = name or "__unnamed__"
	if type(t) ~= "table" then
		return name .. " = " .. basicSerialize(t)
	end
	cart, autoref = "", ""
	addtocart(t, name, indent, nil, nil, depth + 1)
	return cart .. autoref
end;




--Example =
--local text = RealisticUtils.testTable(self.specializations, "specialization list", "|");
--print(text);


addConsoleCommand("mrsetvalue1", "Set the value of RealisticUtils.mrDebugValue1", "consoleCommandSetValue1", RealisticUtils);
RealisticUtils.consoleCommandSetValue1 = function(unusedSelf, value)
	if value then
		RealisticUtils.mrDebugValue1 = value;
		print("RealisticUtils.consoleCommandSetValue1 ------------------------------------------------- value=" .. tostring(value) .. " - number="..tostring(tonumber(value)));
		return "RealisticUtils.consoleCommandSetValue1 - value=" .. tostring(value);
	end
	return "RealisticUtils.consoleCommandSetValue1 - no value typed as a parameter";
end

addConsoleCommand("mrsetvalue2", "Set the value of RealisticUtils.mrDebugValue2", "consoleCommandSetValue2", RealisticUtils);
RealisticUtils.consoleCommandSetValue2 = function(unusedSelf, value)
	if value then
		RealisticUtils.mrDebugValue2 = value;
		print("RealisticUtils.consoleCommandSetValue2 ------------------------------------------------- value=" .. tostring(value) .. " - number="..tostring(tonumber(value)));
		return "RealisticUtils.consoleCommandSetValue2 - value=" .. tostring(value);
	end
	return "RealisticUtils.consoleCommandSetValue2 - no value typed as a parameter";
end

addConsoleCommand("mrsetvalue3", "Set the value of RealisticUtils.mrDebugValue3", "consoleCommandSetValue3", RealisticUtils);
RealisticUtils.consoleCommandSetValue3 = function(unusedSelf, value)
	if value then
		RealisticUtils.mrDebugValue3 = value;	
		
		print("RealisticUtils.consoleCommandSetValue3 -------------------------------------------------- value=" .. tostring(value) .. " - number="..tostring(tonumber(value)));
		return "RealisticUtils.consoleCommandSetValue3 - value=" .. tostring(value);
	end
	return "RealisticUtils.consoleCommandSetValue3 - no value typed as a parameter";
end









